#
# This script is part of TkKasse. You may use, modify and distribute it under the terms
# of the GNU General Public License, Version 2.
# See details at http://www.fsf.org/licenses/gpl.txt
# (C) by Jan Kandziora <tkkasse@users.sf.net> 
#


## Generic server class:
itcl::class Server \
{
	constructor { args  } { eval configure $args } {}
	destructor {}

	method acceptConnection { FD_ARG REMOTE_ADDRESS_ARG REMOTE_PORT_ARG }
	method closeConnection { CONNECTION_ARG }
	method forEachConnection { args }
	method forEachConnectionBut { CONNECTIONS_ARG args }
	method start {}
	method stop {}

	public \
	{
		variable accept_port {}
		variable connection_class {}
	}

	protected \
	{
		variable accept_fd {}
		variable connections {}
	}
}

## Generic server destructor:
itcl::body Server::destructor {} \
{
	## Stop the server.
	stop
}

## Accept a tcp connection request.
## NOTE: This is a callback and not intended as a user callable method.
itcl::body Server::acceptConnection { FD_ARG REMOTE_ADDRESS_ARG REMOTE_PORT_ARG } \
{
	## Make a new connection
	set CONNECTION [ $connection_class #auto $this $FD_ARG $REMOTE_ADDRESS_ARG $REMOTE_PORT_ARG ]
	
	## Append it to the list of active connections.
	lappend connections [ namespace which $CONNECTION ]

	log info "Connection [ namespace which $CONNECTION ] to $REMOTE_ADDRESS_ARG:$REMOTE_PORT_ARG established."
}	

## Close a connection.
## NOTE: This is a callback and not intended as a user callable method.
itcl::body Server::closeConnection { CONNECTION_ARG } \
{
	## The connection was closed by the peer. Delete the connection object.
	itcl::delete object $CONNECTION_ARG

 	## Remove it from the connection list.
	set INDEX [ lsearch -exact $connections $CONNECTION_ARG ]
	set connections [ lreplace $connections $INDEX $INDEX ]

	log info "Connection $CONNECTION_ARG closed."
}

## Evaluate args in the scope of each connection in the server context:
itcl::body Server::forEachConnection { args } \
{
	foreach CONNECTION $connections \
	{
		$CONNECTION eval $args
	}
}

## Evaluate args in the scope of each connection but the mentioned in the server context:
itcl::body Server::forEachConnectionBut { CONNECTIONS_ARG args } \
{
	foreach CONNECTION $connections \
	{
		## Leave out the mentioned connections.
		if { [ lsearch -exact $CONNECTIONS_ARG $CONNECTION ] == -1 } { $CONNECTION eval $args }
	}
}

## Accept tcp connection requests a newly created server socket.
itcl::body Server::start {} \
{
	set accept_fd [ socket -server "$this acceptConnection" $accept_port ]

	log info "Server $this started. Listening on tcp port $accept_port."
}

## Shut down the server.
itcl::body Server::stop {} \
{
	## Do not accept any new connections.
	close $accept_fd

	## Disconnect and destroy all pending connections.
	foreach CONNECTION $connections { itcl::delete object $CONNECTION }

	log info "Server $this stopped."
}



## Authenticated server class:
itcl::class Authenticated::Server \
{
	inherit ::Server

	constructor { args  } { eval configure $args } {}

	method authenticateUser { USER_LOGIN_NAME_ARG USER_PASSWORD_ARG }

	public variable user_manager

}

## Authenticate an user:
itcl::body Authenticated::Server::authenticateUser { USER_LOGIN_NAME_ARG USER_PASSWORD_ARG } \
{
	## Let the user manager make the decision:
	$user_manager authenticateUser $USER_LOGIN_NAME_ARG $USER_PASSWORD_ARG
}

