#
# This script is part of TkKasse. You may use, modify and distribute it under the terms
# of the GNU General Public License, Version 2.
# See details at http://www.fsf.org/licenses/gpl.txt
# (C) by Jan Kandziora <tkkasse@users.sf.net> 
#


## Article class:
itcl::class Article \
{
	constructor { args  } { eval configure $args } {}
	
	public \
	{
		variable key {}
		variable name {}
	}
}



## Article manager class:
itcl::class ArticleManager \
{
	constructor { args  } { eval configure $args } {}

	method addArticle { ARTICLE_OBJECT_ARG }
	method getArticle { KEY_ARG }
	method removeArticle { ARTICLE_OBJECT_ARG }

	public variable articles {}
}

## Add an article:
itcl::body ArticleManager::addArticle { ARTICLE_OBJECT_ARG } \
{
	## Test whether the given key is unique.
	foreach ARTICLE_OBJECT $articles \
	{
		if { [ $ARTICLE_OBJECT_ARG cget -key ] == [ $ARTICLE_OBJECT cget -key ] } \
		{
			## This article already exists in our list.
			return -code error -errorcode { TKKASSE ARTICLE EXISTS } "article already exists"
		}
	}

	## Add the article object descriptor to our list.
	lappend articles [ namespace which $ARTICLE_OBJECT_ARG ]
}

## Get an article by its key:
itcl::body ArticleManager::getArticle { KEY_ARG } \
{
	## Find an article object which matches the template.
	foreach ARTICLE_OBJECT $articles \
	{
		## Is this the article object?
		if { [ $ARTICLE_OBJECT cget -key ] == $KEY_ARG } \
		{
			## Success. Return the article descriptor.
			return $ARTICLE_OBJECT 
		}
	}

	## No such article.
	return -code error -errorcode { TKKASSE ARTICLE UNKNONW } "article unknown."
}

## Remove an article:
itcl::body ArticleManager::removeArticle { ARTICLE_OBJECT_ARG } \
{
	## Find the article object for this key.
	foreach ARTICLE_OBJECT $articles \
	{
		## Is this the article object?
		if { [ $ARTICLE_OBJECT_ARG cget -key ] == [ $ARTICLE_OBJECT cget -key ] } \
		{
			## Yes. Take this article object's descriptor from the list.
			set INDEX [ lsearch -exact $articles $ARTICLE_OBJECT ]
			set articles [ lreplace $articles $INDEX $INDEX ]

			## Return with success.
			return 
		}
	}

	## The article was not found. Throw an error.
	return -code error -errorcode { TKKASSE ARTICLE UNKNOWN } "article unknown."
}



## Evaluate in the TkKasse namespace
namespace eval TkKasse \
{
	# TkKasse article class:
	itcl::class Article \
	{
		inherit ::Article

		constructor { args } { eval configure $args } {}

		public \
		{
			variable price {}
			variable service {}
		}
	}



	# TkKasse article class:
	itcl::class ArticleGroup \
	{
		inherit ::Article

		constructor { args } { eval configure $args } {}
	}



	## TkKasse article manager class:
	itcl::class ArticleManager \
	{
		inherit ::ArticleManager

		constructor { args } { eval configure $args } {}
		method getServices {}
	}

	## Get all services of the articles in the article manager:
	itcl::body ArticleManager::getServices {} \
	{
		set SERVICES {}

		## Go through all articles.
		foreach ARTICLE $articles \
		{
			## Check for an actual article.
			if [ $ARTICLE isa ::TkKasse::Article ] \
			{
				# It is. If the service is not yet listed, add it.
				set SERVICE [ $ARTICLE cget -service ]
				if { [ lsearch -exact $SERVICES $SERVICE ] == -1 } { lappend SERVICES $SERVICE }
			}
		}

		## Return the service list.
		return $SERVICES
	}
}

