#
# This script is part of TkKasse. You may use, modify and distribute it under the terms
# of the GNU General Public License, Version 2.
# See details at http://www.fsf.org/licenses/gpl.txt
# (C) by Jan Kandziora <tkkasse@users.sf.net> 
#



## Evaluate in the TkKasse namespace
namespace eval TkKasse \
{
	## Bill class:
	itcl::class Bill \
	{
		constructor { args } { eval configure $args } {}

		method addArticleItem { KEY_ARG NAME_ARG PRICE_ARG SERVICE_ARG }
		method addArticleItemAgain { KEY_ARG NAME_ARG PRICE_ARG SERVICE_ARG }
		method addExtraToItem { ITEM_UNIQUE_ARG NAME_ARG PRICE_ARG }
		method cancelItem { ITEM_UNIQUE_LIST_ARG }
		method changeOrderOfItem { ITEM_UNIQUE_LIST_ARG }
		method changeServiceOfItem { ITEM_UNIQUE_LIST_ARG SERVICES_ARG }
		method changeTaxOfItem { ITEM_UNIQUE_LIST_ARG }
		method decrementItemMultiplier { ITEM_UNIQUE_LIST_ARG }
		method deliverItem { ITEM_UNIQUE_LIST_ARG }
		method getExtraIndexByUnique { ITEM_INDEX_ARG ITEM_UNIQUE_ARG }
		method getItemIndexByUnique { ITEM_UNIQUE_ARG }
		method incrementItemMultiplier { ITEM_UNIQUE_LIST_ARG }
		method orderItem { ITEM_UNIQUE_LIST_ARG }
		method printGuestReceipt { PRINT_MANAGER_ARG DISPLAY_ARG HOW_ARG }
		method printOrderReceipt { ITEM_UNIQUE_LIST_ARG PRINT_MANAGER_ARG DISPLAY_ARG }
		method removeExtraFromItem { ITEM_UNIQUE_ARG EXTRA_INDEX_ARG }
		method updateSums {}

		public \
		{
			variable cancelled_sum 0
			variable close_time {}
			variable create_time {}
			variable delivered_sum 0
			variable delivered_sum_low_tax 0
			variable delivered_sum_high_tax 0
			variable unique_counter 0
			variable high_tax 0
			variable items {}
			variable low_tax 0
			variable ordered_sum 0
			variable payed_by {}
			variable payed_comment {}
			variable print_time {}
			variable showed_sum 0
			variable table {}
			variable tag {}
			variable user {}
		}
	}

	## Bill class constructor:
	itcl::body Bill::constructor { args } \
	{
		## Fill in the create time if none is given.
		if { $create_time == {} } { set create_time [ clock seconds ] }
	}

	## Add a new article to this bill.
	itcl::body Bill::addArticleItem { KEY_ARG NAME_ARG PRICE_ARG SERVICE_ARG } \
	{
		## Append the item to the item list.
		lappend items [ list -cancel 0 -cancelled 0 -delivered 0 -extras {} -key $KEY_ARG -multiplier 1 -name $NAME_ARG -order 0 -ordered 0 -unique $unique_counter -price $PRICE_ARG -service $SERVICE_ARG -tax 0 ]
		incr unique_counter

		## Update the sums of the bill.
		updateSums 
	}

	## Add a new article to this bill, if its key is different from the last one.
	## Otherwise increment the multiplier of the last one.
	itcl::body Bill::addArticleItemAgain { KEY_ARG NAME_ARG PRICE_ARG SERVICE_ARG } \
	{
		## Get the last item in the list.
		array set LAST_ITEM_ARRAY [ lindex $items end ]
		
		## Check its key and if the last item is neither ordered nor delivered nor cancelled nor cancel.
		if { [ catch { set LAST_ITEM_ARRAY(-key) } ] || ( $LAST_ITEM_ARRAY(-key) != $KEY_ARG ) || $LAST_ITEM_ARRAY(-ordered) || $LAST_ITEM_ARRAY(-delivered) || $LAST_ITEM_ARRAY(-cancelled) || $LAST_ITEM_ARRAY(-cancel) } \
		{
			## The article is the first one or different from the last one, or the last one was ordered etc.
			## Append the new item to the item list.
			lappend items [ list -cancel 0 -cancelled 0 -delivered 0 -extras {} -key $KEY_ARG -multiplier 1 -name $NAME_ARG -order 0 -ordered 0 -unique $unique_counter -price $PRICE_ARG -service $SERVICE_ARG -tax 0 ]
			incr unique_counter
		} \
		{
			## This is the same article type. Increment the multiplier.
			incr LAST_ITEM_ARRAY(-multiplier)

			## Change the last item of the list.
			set items [ lreplace $items end end [ array get LAST_ITEM_ARRAY ] ]
		}

		## Update the sums of the bill.
		updateSums 
	}

	## Add an extra to this items' extra list.
	itcl::body Bill::addExtraToItem { ITEM_UNIQUE_ARG NAME_ARG PRICE_ARG } \
	{
		## Check for the symbolic "end" index.
		if { $ITEM_UNIQUE_ARG == "end" } { set ITEM_INDEX end } \
		{
			## Get the item in the list. Ignore the request if there is no match.
			if [ catch { getItemIndexByUnique $ITEM_UNIQUE_ARG } ITEM_INDEX ] return
		}
		set ITEM [ lindex $items $ITEM_INDEX ]

		## Ignore the request if there is no item.
		if { $ITEM == {} } return

		## Check the item is neither cancel, cancelled, delivered, ordered.
		if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-cancel" ] + 1 ] ] return
		if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-cancelled" ] + 1 ] ] return
		if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-delivered" ] + 1 ] ] return
		if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-ordered" ] + 1 ] ] return

		## Get the item in the list.
		array set ITEM_ARRAY $ITEM

		## Add the extra to the item's extras list.
		lappend ITEM_ARRAY(-extras) [ list -name $NAME_ARG -unique $unique_counter -price $PRICE_ARG ]
		incr unique_counter

		## Change the item.
		set items [ lreplace $items $ITEM_INDEX $ITEM_INDEX [ array get ITEM_ARRAY ] ]

		## Update the sums of the bill.
		updateSums 
	}

	## Cancel an item.
	itcl::body Bill::cancelItem { ITEM_UNIQUE_LIST_ARG } \
	{
		## Process the list of keys.
		foreach ITEM_UNIQUE $ITEM_UNIQUE_LIST_ARG \
		{
			## Get the item in the list. Ignore the request if there is no match.
			if [ catch { getItemIndexByUnique $ITEM_UNIQUE } ITEM_INDEX ] continue
			set ITEM [ lindex $items $ITEM_INDEX ]

			## Check if the item is ordered.
			if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-ordered" ] + 1 ] ] \
			{
				## Yes. Check the item is neither cancel or cancelled.
				if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-cancel" ] + 1 ] ] continue
				if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-cancelled" ] + 1 ] ] continue

				## Make a cancel entry from the given one.
				array set CANCEL_ITEM_ARRAY $ITEM
				set CANCEL_ITEM_ARRAY(-cancel) 1

				## If the item was delivered, also mark the cancel item as ordered and delivered.
				set DELIVERED_STATUS [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-delivered" ] + 1 ] ]
				set CANCEL_ITEM_ARRAY(-delivered) $DELIVERED_STATUS
				set CANCEL_ITEM_ARRAY(-ordered) $DELIVERED_STATUS

				## Set the other fields.
				set CANCEL_ITEM_ARRAY(-price) "-$CANCEL_ITEM_ARRAY(-price)"
				set CANCEL_ITEM_ARRAY(-unique) $unique_counter
				incr unique_counter

				## Add the price to the cancelled_sum field of the bill
				set cancelled_sum [ expr $cancelled_sum + $CANCEL_ITEM_ARRAY(-price) * $CANCEL_ITEM_ARRAY(-multiplier) ]

				## Alter all the associated extras.
				set NEW_EXTRAS {}
				foreach EXTRA $CANCEL_ITEM_ARRAY(-extras) \
				{
					array set EXTRA_ARRAY $EXTRA
					set EXTRA_ARRAY(-price) "-$EXTRA_ARRAY(-price)"
					set EXTRA_ARRAY(-unique) $unique_counter
					incr unique_counter

					## Add the price to the cancelled_sum field of the bill
					set cancelled_sum [ expr $cancelled_sum + $EXTRA_ARRAY(-price) * $CANCEL_ITEM_ARRAY(-multiplier) ]

					lappend NEW_EXTRAS [ array get EXTRA_ARRAY ]
					unset EXTRA_ARRAY
				}
				set CANCEL_ITEM_ARRAY(-extras) $NEW_EXTRAS

				## Append the new cancel entry to the list.
				lappend items [ array get CANCEL_ITEM_ARRAY ]

				## Mark the orgininal item as cancelled.
				set ITEM_FIELD_INDEX [ expr [ lsearch -exact $ITEM "-cancelled" ] + 1 ]
				set items [ lreplace $items $ITEM_INDEX $ITEM_INDEX [ lreplace $ITEM $ITEM_FIELD_INDEX $ITEM_FIELD_INDEX 1 ] ]
			} \
			{
				## No. Check the item is neither cancel or cancelled,.
				if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-cancel" ] + 1 ] ] continue
				if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-cancelled" ] + 1 ] ] continue

				## Remove the item of the list.
				set items [ lreplace $items $ITEM_INDEX $ITEM_INDEX ]
			}
		}

		## Update the sums of the bill.
		updateSums 
	}

	## Change the order counter of an item.
	itcl::body Bill::changeOrderOfItem { ITEM_UNIQUE_LIST_ARG } \
	{
		## Process the list of keys.
		foreach ITEM_UNIQUE $ITEM_UNIQUE_LIST_ARG \
		{
			## Get the item in the list. Ignore the request if there is no match.
			if [ catch { getItemIndexByUnique $ITEM_UNIQUE } ITEM_INDEX ] continue
			set ITEM [ lindex $items $ITEM_INDEX ]
	
			## Check the item is neither cancel, cancelled, delivered, ordered.
			if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-cancel" ] + 1 ] ] continue
			if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-cancelled" ] + 1 ] ] continue
			if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-delivered" ] + 1 ] ] continue
			if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-ordered" ] + 1 ] ] continue

			## Get the order field.
			set ORDER_COUNTER_INDEX [ expr [ lsearch -exact $ITEM "-order" ] + 1 ]
			set ORDER [ lindex $ITEM $ORDER_COUNTER_INDEX ]
		
			## Toggle the order counter.
			set ITEM [ lreplace $ITEM $ORDER_COUNTER_INDEX $ORDER_COUNTER_INDEX [ expr 1 - $ORDER ] ]

			## Change the item of the list.
			set items [ lreplace $items $ITEM_INDEX $ITEM_INDEX $ITEM ]
		}
	
		## Update the sums of the bill.
		updateSums 
	}

	## Change the service of an item.
	itcl::body Bill::changeServiceOfItem { ITEM_UNIQUE_LIST_ARG SERVICES_ARG } \
	{
		## Process the list of keys.
		foreach ITEM_UNIQUE $ITEM_UNIQUE_LIST_ARG \
		{
			## Get the item in the list. Ignore the request if there is no match.
			if [ catch { getItemIndexByUnique $ITEM_UNIQUE } ITEM_INDEX ] continue
			set ITEM [ lindex $items $ITEM_INDEX ]
	
			## Check the item is neither cancel, cancelled, delivered, ordered.
			if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-cancel" ] + 1 ] ] continue
			if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-cancelled" ] + 1 ] ] continue
			if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-delivered" ] + 1 ] ] continue
			if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-ordered" ] + 1 ] ] continue

			## Get the service field.
			set SERVICE_COUNTER_INDEX [ expr [ lsearch -exact $ITEM "-service" ] + 1 ]
			set SERVICE [ lindex $ITEM $SERVICE_COUNTER_INDEX ]
		
			## Change the service.
			set SERVICE_INDEX [ lsearch -exact $SERVICES_ARG $SERVICE ]
			if { $SERVICE_INDEX == [ expr [ llength $SERVICES_ARG ] - 1 ] } { set SERVICE_INDEX 0 } { incr SERVICE_INDEX }
			set ITEM [ lreplace $ITEM $SERVICE_COUNTER_INDEX $SERVICE_COUNTER_INDEX [ lindex $SERVICES_ARG $SERVICE_INDEX ] ]

			## Change the item of the list.
			set items [ lreplace $items $ITEM_INDEX $ITEM_INDEX $ITEM ]
		}
	
		## Update the sums of the bill.
		updateSums 
	}

	## Change the tax of an item.
	itcl::body Bill::changeTaxOfItem { ITEM_UNIQUE_LIST_ARG } \
	{
		## Process the list of keys.
		foreach ITEM_UNIQUE $ITEM_UNIQUE_LIST_ARG \
		{
			## Get the item in the list. Ignore the request if there is no match.
			if [ catch { getItemIndexByUnique $ITEM_UNIQUE } ITEM_INDEX ] continue
			set ITEM [ lindex $items $ITEM_INDEX ]
	
			## Check the item is neither cancel, cancelled, delivered, ordered.
			if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-cancel" ] + 1 ] ] continue
			if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-cancelled" ] + 1 ] ] continue
			if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-delivered" ] + 1 ] ] continue
			if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-ordered" ] + 1 ] ] continue

			## Get the order field.
			set TAX_INDEX [ expr [ lsearch -exact $ITEM "-tax" ] + 1 ]
			set TAX [ lindex $ITEM $TAX_INDEX ]
		
			## Toggle the order counter.
			set ITEM [ lreplace $ITEM $TAX_INDEX $TAX_INDEX [ expr 1 - $TAX ] ]

			## Change the item of the list.
			set items [ lreplace $items $ITEM_INDEX $ITEM_INDEX $ITEM ]
		}
	
		## Update the sums of the bill.
		updateSums 
	}

	## Decrement the multiplier of an item.
	itcl::body Bill::decrementItemMultiplier { ITEM_UNIQUE_LIST_ARG } \
	{
		## Process the list of keys.
		foreach ITEM_UNIQUE $ITEM_UNIQUE_LIST_ARG \
		{
			## Get the item in the list. Ignore the request if there is no match.
			if [ catch { getItemIndexByUnique $ITEM_UNIQUE } ITEM_INDEX ] continue
			set ITEM [ lindex $items $ITEM_INDEX ]
	
			## Check the item is neither cancel, cancelled, delivered, ordered.
			if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-cancel" ] + 1 ] ] continue
			if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-cancelled" ] + 1 ] ] continue
			if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-delivered" ] + 1 ] ] continue
			if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-ordered" ] + 1 ] ] continue

			## Get the multiplier.
			set MULTIPLIER_INDEX [ expr [ lsearch -exact $ITEM "-multiplier" ] + 1 ]
			set MULTIPLIER [ lindex $ITEM $MULTIPLIER_INDEX ]
		
			## Ignore the request if the multiplier is already one.
			if { $MULTIPLIER == 1 } continue

			## Decrement the multiplier.
			set ITEM [ lreplace $ITEM $MULTIPLIER_INDEX $MULTIPLIER_INDEX [ expr $MULTIPLIER - 1 ] ]

			## Change the item of the list.
			set items [ lreplace $items $ITEM_INDEX $ITEM_INDEX $ITEM ]
		}
	
		## Update the sums of the bill.
		updateSums 
	}

	## Deliver an item.
	itcl::body Bill::deliverItem { ITEM_UNIQUE_LIST_ARG } \
	{
		## Process the list of keys.
		foreach ITEM_UNIQUE $ITEM_UNIQUE_LIST_ARG \
		{
			## Get the item in the list. Ignore the request if there is no match.
			if [ catch { getItemIndexByUnique $ITEM_UNIQUE } ITEM_INDEX ] continue
			set ITEM [ lindex $items $ITEM_INDEX ]

			## Check the item is ordered.
			if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-ordered" ] + 1 ] ] \
			{
				## Mark the item as delivered.
				set ITEM_FIELD_INDEX [ expr [ lsearch -exact $ITEM "-delivered" ] + 1 ]
				set ITEM [ lreplace $ITEM $ITEM_FIELD_INDEX $ITEM_FIELD_INDEX 1 ]

				## Change the item in the list.
				set items [ lreplace $items $ITEM_INDEX $ITEM_INDEX $ITEM ]
			}
		}
	
		## Update the sums of the bill.
		updateSums 
	}

	## Get the index of an item by its entry unique.
	## This is done as items can't be removed from the list due to race conditions
	## when two clients access the same bill at the same time.
	## Deleted items are not displayed on the client side, so we need another index
	## to correctly identify the item in question, the unique.
	itcl::body Bill::getItemIndexByUnique { ITEM_UNIQUE_ARG } \
	{
		## Look through all items.
		set INDEX 0
		foreach ITEM $items \
		{
			## Check the unique entry.
			if { [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-unique" ] + 1 ] ] == $ITEM_UNIQUE_ARG } { return $INDEX }

			## Next index.
			incr INDEX
		}

		## Not found. This is an error.
		return -code error -errorcode { TKKASSE BILL NO_SUCH_ITEM } "no such item"
	}

	## Get the index of an extra within an item by its entry unique.
	itcl::body Bill::getExtraIndexByUnique { ITEM_INDEX_ARG EXTRA_UNIQUE_ARG } \
	{
		## Get the extras.
 		set ITEM [ lindex $items $ITEM_INDEX_ARG ]
		set EXTRAS [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-extras" ] + 1 ] ]

		## Look through all extras of this item.
		set INDEX 0
		foreach EXTRA $EXTRAS \
		{
			## Check the unique entry.
			if { [ lindex $EXTRA [ expr [ lsearch -exact $EXTRA "-unique" ] + 1 ] ] == $EXTRA_UNIQUE_ARG } { return $INDEX }

			## Next index.
			incr INDEX
		}

		## Not found. This is an error.
		return -code error -errorcode { TKKASSE BILL NO_SUCH_ITEM_EXTRA } "no such item extra"
	}

	## Increment the Multiplier of an item.
	itcl::body Bill::incrementItemMultiplier { ITEM_UNIQUE_LIST_ARG } \
	{
		## Process the list of keys.
		foreach ITEM_UNIQUE $ITEM_UNIQUE_LIST_ARG \
		{
			## Get the item in the list. Ignore the request if there is no match.
			if [ catch { getItemIndexByUnique $ITEM_UNIQUE } ITEM_INDEX ] continue
			set ITEM [ lindex $items $ITEM_INDEX ]

			## Check the item is neither cancel, cancelled, delivered, ordered.
			if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-cancel" ] + 1 ] ] continue
			if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-cancelled" ] + 1 ] ] continue
			if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-delivered" ] + 1 ] ] continue
			if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-ordered" ] + 1 ] ] continue

			## Increment the multiplier.
			set MULTIPLIER_INDEX [ expr [ lsearch -exact $ITEM "-multiplier" ] + 1 ]
			set ITEM [ lreplace $ITEM $MULTIPLIER_INDEX $MULTIPLIER_INDEX [ expr [ lindex $ITEM $MULTIPLIER_INDEX ] + 1 ] ]

			## Change the item of the list.
			set items [ lreplace $items $ITEM_INDEX $ITEM_INDEX $ITEM ]
		}	

		## Update the sums of the bill.
		updateSums 
	}

	## Order an item.
	itcl::body Bill::orderItem { ITEM_UNIQUE_LIST_ARG } \
	{
		## Process the list of keys.
		foreach ITEM_UNIQUE $ITEM_UNIQUE_LIST_ARG \
		{
			## Get the item in the list. Ignore the request if there is no match.
			if [ catch { getItemIndexByUnique $ITEM_UNIQUE } ITEM_INDEX ] continue 
			set ITEM [ lindex $items $ITEM_INDEX ]

			## Mark the  item as ordered.
			set ITEM_FIELD_INDEX [ expr [ lsearch -exact $ITEM "-ordered" ] + 1 ]
			set ITEM [ lreplace $ITEM $ITEM_FIELD_INDEX $ITEM_FIELD_INDEX 1 ]

			## Change the item in the list.
			set items [ lreplace $items $ITEM_INDEX $ITEM_INDEX $ITEM ]
		}

		## Update the sums of the bill.
		updateSums
	}

	## Print a guest receipt.
	itcl::body Bill::printGuestReceipt { PRINT_MANAGER_ARG DISPLAY_ARG HOW_ARG } \
	{
		## Ensure all the items are delivered.
		foreach ITEM $items \
		{
			if { [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-delivered" ] + 1 ] ] == 0 } \
			{
				return -code error -errorcode { TKKASSE BILL NOT_COMPLETE } "bill not complete"
			}
		}

		## Set the print time, if none is given.
		if { $print_time == {} } { set print_time [ clock seconds ] }

		## Condense all items that are the same apart from multiplier.
		array set ITEMS_ARRAY {}
		foreach ITEM $items \
		{
			catch { unset ITEM_ARRAY }
			array set ITEM_ARRAY $ITEM

			## Skip cancel and cancelled items.
			if $ITEM_ARRAY(-cancel) continue
			if $ITEM_ARRAY(-cancelled) continue

			## Separate multiplier as value, and important other values as key.
			set MULTIPLIER $ITEM_ARRAY(-multiplier)
			unset ITEM_ARRAY(-service) ITEM_ARRAY(-order) ITEM_ARRAY(-unique) ITEM_ARRAY(-multiplier)
			set KEY [ array get ITEM_ARRAY ]

			## Condense items with same keys.
			if [ catch { expr $ITEMS_ARRAY($KEY)+$MULTIPLIER } ITEMS_ARRAY($KEY) ] \
			{
				set ITEMS_ARRAY($KEY) $MULTIPLIER
			}
		}

		## Get the items.
		## Remember the keys of ITEMS_ARRAY are the other important values, and the value is only the multiplier.
		set ITEM_INFO {}
		foreach ITEM [ array names ITEMS_ARRAY ] \
		{
			catch { unset ITEM_ARRAY }
			array set ITEM_ARRAY $ITEM

			## Set up the extras.
			set EXTRAS {}
			set EXTRAS_PRICE 0
			foreach EXTRA $ITEM_ARRAY(-extras) \
			{
				set EXTRA_PRICE [ lindex $EXTRA [ expr [ lsearch -exact $EXTRA "-price" ] + 1 ] ]
				lappend EXTRAS [ list [ lindex $EXTRA [ expr [ lsearch -exact $EXTRA "-name" ] + 1 ] ] $EXTRA_PRICE ]
				set EXTRAS_PRICE [ expr $EXTRAS_PRICE + $EXTRA_PRICE ]
			}

			## Save the item info.
			lappend ITEM_INFO [ list $ITEM_ARRAY(-key) $ITEM_ARRAY(-name) $ITEM_ARRAY(-price) $EXTRAS $ITEMS_ARRAY($ITEM) ]

		}

		## Determine the filter to use for print.
		switch $HOW_ARG \
		{
			plain { set FILTER GuestReceipt }
			spending { set FILTER GuestReceiptSpending }
			room  { set FILTER GuestReceiptRoom }
		}

		## Print the receipt.
		$PRINT_MANAGER_ARG print [ list TIME $print_time \
						TABLE $table \
						TAG $tag \
						ITEMS $ITEM_INFO \
						GROSS_SUM_LOW_TAX $delivered_sum_low_tax \
						TAX_RATE_LOW [ msgcat::mc <tax.rate.low> ] \
						LOW_TAX $low_tax \
						GROSS_SUM_HIGH_TAX $delivered_sum_high_tax \
						TAX_RATE_HIGH [ msgcat::mc <tax.rate.high> ] \
						HIGH_TAX $high_tax \
					] $FILTER $DISPLAY_ARG bar
	}

	## Print an order receipt.
	itcl::body Bill::printOrderReceipt { ITEM_UNIQUE_LIST_ARG PRINT_MANAGER_ARG DISPLAY_ARG } \
	{
		## Process the list of keys.
		array set ITEM_INFO {}
		foreach ITEM_UNIQUE $ITEM_UNIQUE_LIST_ARG \
		{
			## Get the item in the list. Ignore the request if there is no match.
			if [ catch { getItemIndexByUnique $ITEM_UNIQUE } ITEM_INDEX ] continue 
			set ITEM [ lindex $items $ITEM_INDEX ]

			## Get the item data.
			catch { unset ITEM_ARRAY }
			array set ITEM_ARRAY $ITEM

			## Ignore already ordered items.
			if $ITEM_ARRAY(-ordered) continue

			## Set up the extras.
			set EXTRAS {}
			foreach EXTRA $ITEM_ARRAY(-extras) \
			{
				lappend EXTRAS [ lindex $EXTRA [ expr [ lsearch -exact $EXTRA "-name" ] + 1 ] ]
			}

			## Save the item info.
			lappend ITEM_INFO($ITEM_ARRAY(-service)) [ list $ITEM_ARRAY(-key) $ITEM_ARRAY(-name) $EXTRAS $ITEM_ARRAY(-multiplier) $ITEM_ARRAY(-cancel) ]
		}

		## Print a receipt for each service with orders.
		foreach SERVICE [ array names ITEM_INFO ] \
		{
			$PRINT_MANAGER_ARG print [ list TABLE $table TAG $tag ITEMS $ITEM_INFO($SERVICE) ] OrderReceipt $DISPLAY_ARG $SERVICE 
		}
	}

	## Remove an extra from this articles' extra list.
	itcl::body Bill::removeExtraFromItem { ITEM_UNIQUE_ARG EXTRA_UNIQUE_ARG } \
	{
		## Get the item in the list. Ignore the request if there is no match.
		if [ catch { getItemIndexByUnique $ITEM_UNIQUE_ARG } ITEM_INDEX ] return
		set ITEM [ lindex $items $ITEM_INDEX ]

		## Check the item is neither cancel, cancelled, delivered, ordered.
		if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-cancel" ] + 1 ] ] return
		if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-cancelled" ] + 1 ] ] return
		if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-delivered" ] + 1 ] ] return
		if [ lindex $ITEM [ expr [ lsearch -exact $ITEM "-ordered" ] + 1 ] ] return

		## Get the extras field. Ignore the request if there is no match.
		set EXTRAS_FIELD_INDEX [ expr [ lsearch -exact $ITEM "-extras" ] + 1 ]
		set EXTRAS [ lindex $ITEM $EXTRAS_FIELD_INDEX ]
 		if [ catch { getExtraIndexByUnique $ITEM_INDEX $EXTRA_UNIQUE_ARG } EXTRA_INDEX ] return

		## Change the extra record in the item. Remove the extra from it. 
		set ITEM [ lreplace $ITEM $EXTRAS_FIELD_INDEX $EXTRAS_FIELD_INDEX [ lreplace $EXTRAS $EXTRA_INDEX $EXTRA_INDEX ] ]

		## Change the item of the list.
		set items [ lreplace $items $ITEM_INDEX $ITEM_INDEX $ITEM ]
	
		## Update the sums of the bill.
		updateSums 
	}

	## Update all the sums of the bill.
	itcl::body Bill::updateSums {} \
	{
		## Reset all sums.
		set showed_sum 0
		set ordered_sum 0
		set delivered_sum 0
		set delivered_sum_low_tax 0
		set delivered_sum_high_tax 0
		set low_tax 0
		set high_tax 0

		## Go through all items.
		foreach ITEM $items \
		{	
			catch { unset ITEM_ARRAY }
			array set ITEM_ARRAY $ITEM

			## Add the price * multiplier to the showed sum. 
			set showed_sum [ expr $showed_sum  + $ITEM_ARRAY(-price) * $ITEM_ARRAY(-multiplier) ]
			
			## If this item is ordered, also add it to the ordered sum.
			if $ITEM_ARRAY(-ordered) { set ordered_sum [ expr $ordered_sum  + $ITEM_ARRAY(-price) * $ITEM_ARRAY(-multiplier) ] }

			## If this item is delivered, also add it to the delivered sum.
			if $ITEM_ARRAY(-delivered) \
			{
				set delivered_sum [ expr $delivered_sum  + $ITEM_ARRAY(-price) * $ITEM_ARRAY(-multiplier) ]
				if $ITEM_ARRAY(-tax) \
				{
					set delivered_sum_low_tax [ expr $delivered_sum_low_tax + $ITEM_ARRAY(-price) * $ITEM_ARRAY(-multiplier) ]
				} \
				{
					set delivered_sum_high_tax [ expr $delivered_sum_high_tax + $ITEM_ARRAY(-price) * $ITEM_ARRAY(-multiplier) ]
				}
			}

			## Go through all extras of this item.
			foreach EXTRA $ITEM_ARRAY(-extras) \
			{
				set EXTRA_PRICE [ expr [ lindex $EXTRA [ expr [ lsearch -exact $EXTRA "-price" ] + 1 ] ] * $ITEM_ARRAY(-multiplier) ]

				## Add the price * multiplier of each extra to the showed sum. 
				set showed_sum [ expr $showed_sum + $EXTRA_PRICE ]
		
				## If this item is ordered, also add the price * multiplier of the extra to the ordered sum. 
				if $ITEM_ARRAY(-ordered) { set ordered_sum [ expr $ordered_sum + $EXTRA_PRICE ] }
				
				## If this item is delivered, also add the price * multiplier of the extra to the delivered sum. 
				if $ITEM_ARRAY(-delivered) \
				{
					set delivered_sum [ expr $delivered_sum + $EXTRA_PRICE ]
					if $ITEM_ARRAY(-tax) \
					{
						set delivered_sum_low_tax [ expr $delivered_sum_low_tax  + $EXTRA_PRICE ]
					} \
					{
						set delivered_sum_high_tax [ expr $delivered_sum_high_tax + $EXTRA_PRICE ]
					}
				}
			}
		}

		## Calculate tax.
		set LOW_TAX_10 [ expr ($delivered_sum_low_tax * 10 * [ msgcat::mc <tax.rate.low> ]) / (100 + [ msgcat::mc <tax.rate.low> ]) ]
		switch -regexp [ string index $LOW_TAX_10 end ] \
		{
			[0-4] { set low_tax [ expr $LOW_TAX_10 / 10 ] }
			[5-9] { set low_tax [ expr $LOW_TAX_10 / 10 + 1  ] }
		}

		set HIGH_TAX_10 [ expr ($delivered_sum_high_tax * 10 * [ msgcat::mc <tax.rate.high> ]) / (100 + [ msgcat::mc <tax.rate.high> ]) ]
		switch -regexp [ string index $HIGH_TAX_10 end ] \
		{
			[0-4] { set high_tax [ expr $HIGH_TAX_10 / 10 ] }
			[5-9] { set high_tax [ expr $HIGH_TAX_10 / 10 + 1 ] }
		}

	}



	## Cash register class:
	itcl::class CashRegister \
	{
		constructor { args } { eval configure $args } {}
		destructor {}
	
		method addBill { BILL_OBJECT_ARG }
		method cleanUp {}
		method closeBill { BILL_OBJECT_ARG }
		method getArticleByKey { KEY_ARG }
		method getArticleList {}
		method getBill { TAG_ARG }
		method getServices {}
		method incrementSerialNumberCounter {}
		method printReceipt { PRINT_MANAGER_ARG DISPLAY_ARG }
		method printWaiterReceipt { USER_ARG PRINT_MANAGER_ARG DISPLAY_ARG }

		public \
		{
			variable article_manager {}
			variable bills {}
			variable closed_bills {}
			variable serial_number_counter {}
		}
	}

	## Cash register class destructor.
	itcl::body CashRegister::destructor {} \
	{
		## Destroy all associated bill objects.
		foreach BILL [ concat $bills $closed_bills ] { itcl::delete object $BILL }
	}

	## Add a bill.
	itcl::body CashRegister::addBill { BILL_OBJECT_ARG } \
	{
		## Test whether the given key and table is unique.
		foreach BILL_OBJECT $bills \
		{
			if { [ $BILL_OBJECT_ARG cget -tag ] == [ $BILL_OBJECT cget -tag ] } \
			{
				## This bill already exists in our list.
				return -code error -errorcode { TKKASSE BILL EXISTS } "bill already exists"
			}

			if { [ $BILL_OBJECT_ARG cget -table ] == [ $BILL_OBJECT cget -table ] } \
			{
				## This bill already exists in our list.
				return -code error -errorcode { TKKASSE BILL TABLE_OCCUPIED } "table already occupied"
			}
		}

		## Ensure that the tag of this bill is empty.
		if { [ $BILL_OBJECT_ARG cget -tag ] != {} } \
		{
			return -code error -errorcode { TKKASSE BILL HAS_SERIAL } "bill already has a serial number"
		}

		## The bill has no serial number. Assign one depending on the cash register.
		## Increment the serial number in the cash register.
		$BILL_OBJECT_ARG configure -tag $serial_number_counter
		incrementSerialNumberCounter

		## Add the article object descriptor to our list.
		lappend bills [ namespace which $BILL_OBJECT_ARG ]
	}

	## Clean up the cash register.
	itcl::body CashRegister::cleanUp {} \
	{
		log debug "Cleaning up cash register..."

		## Backup before cleaning up.
		backupCashRegister [ lindex [ split $serial_number_counter / ] 0 ]

		## Remove all closed bills.
		foreach BILL_OBJECT $closed_bills \
		{
			itcl::delete object $BILL_OBJECT
		}

		set closed_bills {}

		log debug "All closed bills deleted."

		## Reset the serial number counter.
		set serial_number_counter "[ clock seconds ]/1"
	}

	## Close a bill.
	itcl::body CashRegister::closeBill { BILL_OBJECT_ARG } \
	{
		## Find the bill object for this key.
		foreach BILL_OBJECT $bills \
		{
			## Is this the bill object?
			if { $BILL_OBJECT == $BILL_OBJECT_ARG } \
			{
				## Yes. Is it printed yet?
				if { [ $BILL_OBJECT cget -print_time ] == {} } \
				{
					## No. Throw an error.
					return -code error -errorcode { TKKASSE BILL NOT_PRINTED } "bill not printed yet."
				}

				## Yes. Only for bills not payed in cash: Is the there a comment given? 
				if { [ $BILL_OBJECT cget -payed_by ] != "cash" } \
				{
					if { [ $BILL_OBJECT cget -payed_comment ] == {} } \
					{
						## No. Throw an error.
						return -code error -errorcode { TKKASSE BILL NO_PAY_COMMENT } "no comment given."
					} 
				}

				## Yes. Take this bill object's descriptor from the list.
				set INDEX [ lsearch -exact $bills $BILL_OBJECT ]
				set bills [ lreplace $bills $INDEX $INDEX ]

				## Put the bill on the closed bill list.
				lappend closed_bills $BILL_OBJECT_ARG

				## Return with success.
				return 
			}
		}

		## The bill was not found. Throw an error.
		return -code error -errorcode { TKKASSE BILL UNKNOWN } "bill unknown."
	}

	## Get an article object desciptor for this cash register by its key.
	itcl::body CashRegister::getArticleByKey { KEY_ARG } \
	{
		$article_manager getArticle $KEY_ARG
	}

	## Get article information for this cash register.
	itcl::body CashRegister::getArticleList {} \
	{
		$article_manager cget -articles
	}

	## Get a bill by its tag.
	itcl::body CashRegister::getBill { TAG_ARG } \
	{
		## Find a bill object with the desired tag.
		foreach BILL_OBJECT $bills \
		{
			## Is this the bill object?
			if { [ $BILL_OBJECT cget -tag ] == $TAG_ARG } \
			{
				## Success. Return the bill descriptor.
				return $BILL_OBJECT 
			}
		}

		## No such bill.
		return -code error -errorcode { TKKASSE BILL UNKNOWN } "bill unknown."
	}

	## Get the services for this cash register.
	itcl::body CashRegister::getServices {} \
	{
		$article_manager getServices
	}

	## Increment the serial number counter for added bills.
	itcl::body CashRegister::incrementSerialNumberCounter {} \
	{
		## Get the last complete number in the counter string
		regexp {[0-9]+$} $serial_number_counter NUMBER

		## Increment the number.
		incr NUMBER

		## Replace the last complete number in the counter string by the incremented number
		regsub {(.*?)[0-9]+$} $serial_number_counter "\\1$NUMBER" serial_number_counter
	}

	## Print a receipt of the cash register contents.
	itcl::body CashRegister::printReceipt { PRINT_MANAGER_ARG DISPLAY_ARG } \
	{
		## If there are no closed bills, give an informative message.
		if { $closed_bills == {} } \
		{
			return -code error -errorcode { TKKASSE BILL NO_MATCH } "No bills applicable for this request."
		}

		## Process all closed bills in the cash register.
		set BILLS {}
		set START_TIME {}
		set END_TIME {}
		array set SUMS {}
		set TAX 0
		foreach BILL_OBJECT $closed_bills \
		{
			## Calculate the cash register's working start time.
			if { ( $START_TIME == {} ) || ( $START_TIME > [ $BILL_OBJECT cget -create_time ] ) } { set START_TIME [ $BILL_OBJECT cget -create_time ] }

			## Calculate the cash register's working end time.
			if { ( $END_TIME == {} ) || ( $END_TIME < [ $BILL_OBJECT cget -close_time ] ) } { set END_TIME [ $BILL_OBJECT cget -close_time ] }

			## Get some values from the bill.
			set PAYED_BY [ $BILL_OBJECT cget -payed_by  ]
			set DELIVERED_SUM [ $BILL_OBJECT cget -delivered_sum ]
			set TAX [ expr $TAX + [ $BILL_OBJECT cget -low_tax ] + [ $BILL_OBJECT cget -high_tax ] ]

			## Append the bill description.
			lappend BILLS [ list [ $BILL_OBJECT cget -tag ] $DELIVERED_SUM $PAYED_BY [ $BILL_OBJECT cget -payed_comment ] [ [ $BILL_OBJECT cget -user ] cget -shortname ] [ $BILL_OBJECT cget -cancelled_sum ] ]

			## Add the sum of this bill to the waiter's sums.
			if [ catch { expr $SUMS($PAYED_BY) + $DELIVERED_SUM } SUMS($PAYED_BY) ] { set SUMS($PAYED_BY) $DELIVERED_SUM }
		}

		## Print the cahs register receipt.
		$PRINT_MANAGER_ARG print [ list START_TIME $START_TIME END_TIME $END_TIME BILLS $BILLS SUMS [ array get SUMS ] TAX $TAX ] CashRegisterReceipt $DISPLAY_ARG bar
	}

	## Print a receipt of the cash register contents for a single waiter.
	itcl::body CashRegister::printWaiterReceipt { USER_ARG PRINT_MANAGER_ARG DISPLAY_ARG } \
	{
		## Process all closed bills in the cash register.
		set BILLS {}
		set START_TIME {}
		set END_TIME {}
		array set SUMS {}
		foreach BILL_OBJECT $closed_bills \
		{
			## Check the bill belongs to this user.
			if { [ $BILL_OBJECT cget -user ] == $USER_ARG } \
			{
				## Calculate the waiter's working start time.
				if { ( $START_TIME == {} ) || ( $START_TIME > [ $BILL_OBJECT cget -create_time ] ) } { set START_TIME [ $BILL_OBJECT cget -create_time ] }

				## Calculate the waiter's working end time.
				if { ( $END_TIME == {} ) || ( $END_TIME < [ $BILL_OBJECT cget -close_time ] ) } { set END_TIME [ $BILL_OBJECT cget -close_time ] }

				## Get some values from the bill.
				set PAYED_BY [ $BILL_OBJECT cget -payed_by  ]
				set DELIVERED_SUM [ $BILL_OBJECT cget -delivered_sum ]

				## Append the bill description.
				lappend BILLS [ list [ $BILL_OBJECT cget -tag ] $DELIVERED_SUM $PAYED_BY [ $BILL_OBJECT cget -payed_comment ] [ $BILL_OBJECT cget -cancelled_sum ] ]

				## Add the sum of this bill to the waiter's sums.
				if [ catch { expr $SUMS($PAYED_BY) + $DELIVERED_SUM } SUMS($PAYED_BY) ] { set SUMS($PAYED_BY) $DELIVERED_SUM }
			}
		}

		## If there are no closed bills for this user, give an informative message.
		if { $BILLS == {} } \
		{
			return -code error -errorcode { TKKASSE BILL NO_MATCH } "No bills applicable for this request."
		}

		## Print the waiter receipt.
		$PRINT_MANAGER_ARG print [ list USER [ $USER_ARG cget -name ] START_TIME $START_TIME END_TIME $END_TIME BILLS $BILLS SUMS [ array get SUMS ] ] WaiterReceipt $DISPLAY_ARG bar
	}
}

