#
# This script is part of TkKasse. You may use, modify and distribute it under the terms
# of the GNU General Public License, Version 2.
# See details at http://www.fsf.org/licenses/gpl.txt
# (C) by Jan Kandziora <tkkasse@users.sf.net> 
#


## Evaluate in the TkKasse namespace
namespace eval TkKasse \
{
	## TkKasse connection class:
	itcl::class Connection \
	{
		inherit ::Authenticated::Connection

		constructor { SERVER_ARG FD_ARG REMOTE_HOST_ARG REMOTE_PORT_ARG } { chain $SERVER_ARG $FD_ARG $REMOTE_HOST_ARG $REMOTE_PORT_ARG } {}

		method addArticleItemToBill { TAG_ARG KEY_ARG }
		method addArticleItemToBillAgain { TAG_ARG KEY_ARG }
		method addExtraToBillItem { TAG_ARG ITEM_UNIQUE_LIST_ARG NAME_ARG PRICE_ARG }
		method authenticateUser { USER_LOGIN_NAME_ARG USER_PASSWORD_ARG }
		method cancelItemFromBill { TAG_ARG ITEM_UNIQUE_LIST_ARG }
		method changeOrderOfItemFromBill { TAG_ARG ITEM_UNIQUE_LIST_ARG }
		method changeServiceOfItemFromBill { TAG_ARG ITEM_UNIQUE_LIST_ARG }
		method changeTaxOfItemFromBill { TAG_ARG ITEM_UNIQUE_LIST_ARG }
		method cleanUpCashRegister {}
		method closeBill { TAG_ARG PAYED_BY_ARG PAYED_COMMENT_ARG }
		method createBill { TABLE_ARG }
		method decrementBillItemMultiplier { TAG_ARG ITEM_UNIQUE_LIST_ARG }
		method deliverItemFromBill { TAG_ARG ITEM_UNIQUE_LIST_ARG }
		method doEcho { args }
		method incrementBillItemMultiplier { TAG_ARG ITEM_UNIQUE_LIST_ARG }
		method orderItemFromBill { TAG_ARG ITEM_UNIQUE_LIST_ARG }
		method printCashRegisterReceipt {}
		method printGuestReceiptForBill { TAG_ARG HOW_ARG }
		method printWaiterReceipt {}
		method removeExtraFromBillItem { TAG_ARG ITEM_UNIQUE_LIST_ARG EXTRA_INDEX_ARG }
		method setClientDisplay { DISPLAY_ARG }
		method setArticlesUpdateHandler { HANDLER_ARG }
		method setBillsUpdateHandler { HANDLER_ARG }
		method setBillUpdateHandler { HANDLER_ARG }
		method setLoginHandler { HANDLER_ARG }
		method setUsersUpdateHandler { HANDLER_ARG }
		method updateArticles {}
		method updateBill { TAG_ARG }
		method updateBills { SEE_TAG_ARG }
		method updateUsers {}
	
		protected \
		{
			variable articles_update_handler {}
			variable bills_update_handler {}
			variable bill_update_handler {}
			variable cash_register {}
			variable client_display {} 
			variable login_handler {}
			variable users_update_handler {}
		}
	}

	## TkKasse connection constructor:
	itcl::body Connection::constructor { SERVER_ARG FD_ARG REMOTE_HOST_ARG REMOTE_PORT_ARG } \
	{
		## Get the cash register for this connection.
		set cash_register [ $server cget -cash_register ]

		## Create command aliases.
		$interpreter alias addArticleItemToBill $this addArticleItemToBill
		$interpreter alias addArticleItemToBillAgain $this addArticleItemToBillAgain
		$interpreter alias addExtraToBillItem $this addExtraToBillItem
		$interpreter alias cancelItemFromBill $this cancelItemFromBill
		$interpreter alias changeOrderOfItemFromBill $this changeOrderOfItemFromBill
		$interpreter alias changeServiceOfItemFromBill $this changeServiceOfItemFromBill
		$interpreter alias changeTaxOfItemFromBill $this changeTaxOfItemFromBill
		$interpreter alias cleanUpCashRegister $this cleanUpCashRegister
		$interpreter alias closeBill $this closeBill
		$interpreter alias createBill $this createBill
		$interpreter alias decrementBillItemMultiplier $this decrementBillItemMultiplier
		$interpreter alias deliverItemFromBill $this deliverItemFromBill
		$interpreter alias doEcho $this doEcho
		$interpreter alias incrementBillItemMultiplier $this incrementBillItemMultiplier
		$interpreter alias orderItemFromBill $this orderItemFromBill
		$interpreter alias printCashRegisterReceipt $this printCashRegisterReceipt
		$interpreter alias printGuestReceiptForBill $this printGuestReceiptForBill
		$interpreter alias printWaiterReceipt $this printWaiterReceipt
		$interpreter alias removeExtraFromBillItem $this removeExtraFromBillItem
		$interpreter alias setClientDisplay $this setClientDisplay
		$interpreter alias setArticlesUpdateHandler $this setArticlesUpdateHandler
		$interpreter alias setBillsUpdateHandler $this setBillsUpdateHandler
		$interpreter alias setBillUpdateHandler $this setBillUpdateHandler
		$interpreter alias setLoginHandler $this setLoginHandler
		$interpreter alias setUsersUpdateHandler $this setUsersUpdateHandler
		$interpreter alias updateArticles $this updateArticles
		$interpreter alias updateBills $this updateBills
		$interpreter alias updateUsers $this updateUsers
	}

	## Add an article item to a bill.
	itcl::body Connection::addArticleItemToBill { TAG_ARG KEY_ARG } \
	{
		## Get the bill object descriptor.
		set BILL [ $cash_register getBill $TAG_ARG ]

		## Get the article object descriptor.
		set ARTICLE [ $cash_register getArticleByKey $KEY_ARG ] 

		## Ignore request if this is a TkKasse::articleGroup.
		if [ $ARTICLE isa ::TkKasse::ArticleGroup ] return

		## Add the article to this bill object.
		$BILL addArticleItem [ $ARTICLE cget -key ] [ $ARTICLE cget -name ] [ $ARTICLE cget -price ] [ $ARTICLE cget -service ]

		## Update all the clients.
		$server forEachConnection updateBill $BILL
	}

	## Add an article item to a bill again.
	itcl::body Connection::addArticleItemToBillAgain { TAG_ARG KEY_ARG } \
	{
		## Get the bill object descriptor.
		set BILL [ $cash_register getBill $TAG_ARG ]

		## Get the article object descriptor.
		set ARTICLE [ $cash_register getArticleByKey $KEY_ARG ] 

		## Ignore request if this is a TkKasse::articleGroup.
		if [ $ARTICLE isa ::TkKasse::ArticleGroup ] return

		## Add the article to this bill object.
		$BILL addArticleItemAgain [ $ARTICLE cget -key ] [ $ARTICLE cget -name ] [ $ARTICLE cget -price ] [ $ARTICLE cget -service ]

		## Update all the clients.
		$server forEachConnection updateBill $BILL
	}

	## Add an extra to a bill item.
	itcl::body Connection::addExtraToBillItem { TAG_ARG ITEM_UNIQUE_LIST_ARG NAME_ARG PRICE_ARG } \
	{
		## Get the bill object descriptor.
		set BILL [ $cash_register getBill $TAG_ARG ]

		## Add the article to this bill object.
		$BILL addExtraToItem $ITEM_UNIQUE_LIST_ARG $NAME_ARG $PRICE_ARG 

		## Update all the clients.
		$server forEachConnection updateBill $BILL
	}

	## Authenticate an user:
	itcl::body Connection::authenticateUser { USER_LOGIN_NAME_ARG USER_PASSWORD_ARG } \
	{
		## Authenticate the user.
		chain $USER_LOGIN_NAME_ARG $USER_PASSWORD_ARG

		## Now the user is authenticated.
		## If there is a login handler set, send it.
		if { $login_handler != {} } \
		{
			## Get the configuration of the user object.
			set USER_INFO [ getObjectConfiguration $user ]

			## Remove the password field.
 			set PASSWORD_INDEX [ lsearch -exact $USER_INFO "-password" ]
			set USER_INFO [ lreplace $USER_INFO $PASSWORD_INDEX [ expr $PASSWORD_INDEX +1 ] ]

			## Do some substitutions with the handler and send the result to the client.
			regsub -all {\\|&} [ list $USER_INFO ] {\\\0} SUBST
			regsub -all "%u" $login_handler $SUBST COMMAND
			puts $fd $COMMAND
		} 
	}

	## Cancel an item from a bill.
	itcl::body Connection::cancelItemFromBill { TAG_ARG ITEM_UNIQUE_LIST_ARG } \
	{
		## Get the bill object descriptor.
		set BILL [ $cash_register getBill $TAG_ARG ]

		## Actually cancel the item.
		$BILL cancelItem $ITEM_UNIQUE_LIST_ARG

		## Update all the clients.
		$server forEachConnection updateBill $BILL
	}

	## Change the order counter of an item from a bill.
	itcl::body Connection::changeOrderOfItemFromBill { TAG_ARG ITEM_UNIQUE_LIST_ARG } \
	{
		## Get the bill object descriptor.
		set BILL [ $cash_register getBill $TAG_ARG ]

		## Actually change the order counter.
		$BILL changeOrderOfItem $ITEM_UNIQUE_LIST_ARG

		## Update all the clients.
		$server forEachConnection updateBill $BILL
	}

	## Change the service of an item from a bill.
	itcl::body Connection::changeServiceOfItemFromBill { TAG_ARG ITEM_UNIQUE_LIST_ARG } \
	{
		## Get the bill object descriptor.
		set BILL [ $cash_register getBill $TAG_ARG ]

		## Actually change the service.
		$BILL changeServiceOfItem $ITEM_UNIQUE_LIST_ARG [ $cash_register getServices ]

		## Update all the clients.
		$server forEachConnection updateBill $BILL
	}

	## Change the tax of an item from a bill.
	itcl::body Connection::changeTaxOfItemFromBill { TAG_ARG ITEM_UNIQUE_LIST_ARG } \
	{
		## Get the bill object descriptor.
		set BILL [ $cash_register getBill $TAG_ARG ]

		## Actually change the tax.
		$BILL changeTaxOfItem $ITEM_UNIQUE_LIST_ARG

		## Update all the clients.
		$server forEachConnection updateBill $BILL
	}

	## Clean up the cash register.
	itcl::body Connection::cleanUpCashRegister {} \
	{
		## Test rights.
		testUserRights cleanUp

		## Clean up the cash register.
		$cash_register cleanUp

		## Update all the clients.
		$server forEachConnection updateBills {}
	}

	## Close a bill in the cash register
	itcl::body Connection::closeBill { TAG_ARG PAYED_BY_ARG PAYED_COMMENT_ARG } \
	{
		## Get the bill object descriptor.
		set BILL [ $cash_register getBill $TAG_ARG ]

		## Configure this bill object.
		$BILL configure \
			-close_time [ clock seconds ] \
			-payed_by $PAYED_BY_ARG \
			-payed_comment $PAYED_COMMENT_ARG 

		## Close this bill.
		$cash_register closeBill $BILL

		## Update all the clients.
		$server forEachConnection updateBills {}
	}

	## Create a new bill in the cash register.
	itcl::body Connection::createBill { TABLE_ARG } \
	{
		## Make a new bill object in the global namespace.
		set BILL [ namespace eval :: "::TkKasse::Bill #auto -user $user -table $TABLE_ARG" ]

		## Add it to the cash register.
		if [ catch { $cash_register addBill $BILL } RESULT ] \
		{
			## Adding failed. Destroy the bill object.
			itcl::delete object $BILL
		
			## Report the error
			global errorCode
			global errorInfo
			return -code error -errorcode $errorCode -errorinfo $errorInfo $RESULT
		}

		## Update all the clients.
		$server forEachConnectionBut $this updateBills {}
		updateBills [ $BILL cget -tag ]
	}

	## Decrement the multiplier of the bill.
	itcl::body Connection::decrementBillItemMultiplier { TAG_ARG ITEM_UNIQUE_LIST_ARG } \
	{
		## Get the bill object descriptor.
		set BILL [ $cash_register getBill $TAG_ARG ]

		## Actually decrement the multiplier.
		$BILL decrementItemMultiplier $ITEM_UNIQUE_LIST_ARG

		## Update all the clients.
		$server forEachConnection updateBill $BILL
	}

	## Deliver an item from a bill.
	itcl::body Connection::deliverItemFromBill { TAG_ARG ITEM_UNIQUE_LIST_ARG } \
	{
		## Get the bill object descriptor.
		set BILL [ $cash_register getBill $TAG_ARG ]

		## Actually deliver the item.
		$BILL deliverItem $ITEM_UNIQUE_LIST_ARG

		## Update all the clients.
		$server forEachConnection updateBill $BILL
	}

	## Do an echo:
	itcl::body Connection::doEcho { args } \
	{
		## Send the args back to the client.
		puts $fd $args
	}

	## Increment the multiplier of the bill.
	itcl::body Connection::incrementBillItemMultiplier { TAG_ARG ITEM_UNIQUE_LIST_ARG } \
	{
		## Get the bill object descriptor.
		set BILL [ $cash_register getBill $TAG_ARG ]

		## Actually decrement the multiplier.
		$BILL incrementItemMultiplier $ITEM_UNIQUE_LIST_ARG

		## Update all the clients.
		$server forEachConnection updateBill $BILL
	}

	## Order an item from a bill.
	itcl::body Connection::orderItemFromBill { TAG_ARG ITEM_UNIQUE_LIST_ARG } \
	{
		## Get the bill object descriptor.
		set BILL [ $cash_register getBill $TAG_ARG ]

		## Print an order receipt.
		$BILL printOrderReceipt $ITEM_UNIQUE_LIST_ARG [ $server cget -print_manager ] $client_display

		## Actually order the item.
		$BILL orderItem $ITEM_UNIQUE_LIST_ARG

		## Update all the clients.
		$server forEachConnection updateBill $BILL
	}

	## Print a receipt of all bills in the cash register.
	itcl::body Connection::printCashRegisterReceipt {} \
	{
		## Test rights.
		testUserRights printCashRegisterReceipt

		## Print receipt.
		$cash_register printReceipt [ $server cget -print_manager ] $client_display
	}

	## Print a receipt for a single bill.
	itcl::body Connection::printGuestReceiptForBill { TAG_ARG HOW_ARG } \
	{
		## Get the bill object descriptor.
		set BILL [ $cash_register getBill $TAG_ARG ]

		## Print a guest receipt.
		$BILL printGuestReceipt [ $server cget -print_manager ] $client_display $HOW_ARG
	}

	## Print a receipt of all bills of a single waiter.
	itcl::body Connection::printWaiterReceipt {} \
	{
		$cash_register printWaiterReceipt $user [ $server cget -print_manager ] $client_display
	}

	## Remove an article extra from a bill.
	itcl::body Connection::removeExtraFromBillItem { TAG_ARG ITEM_UNIQUE_LIST_ARG EXTRA_INDEX_ARG } \
	{
		## Get the bill object descriptor.
		set BILL [ $cash_register getBill $TAG_ARG ]

		## Actually remove the extra.
		$BILL removeExtraFromItem $ITEM_UNIQUE_LIST_ARG $EXTRA_INDEX_ARG

		## Update all the clients.
		$server forEachConnection updateBill $BILL
	}

	## Set the client display:
	itcl::body Connection::setClientDisplay { DISPLAY_ARG } \
	{
		set client_display $DISPLAY_ARG
	}

	## Set the article update handler:
	itcl::body Connection::setArticlesUpdateHandler { HANDLER_ARG } \
 	{
		## Set the handler.
		set articles_update_handler $HANDLER_ARG

		## Update the article information on the client of this connection,
		## if there was an handler defined.
		if { $articles_update_handler != {} } updateArticles
	}

	## Set the bills update handler:
	itcl::body Connection::setBillsUpdateHandler { HANDLER_ARG } \
 	{
		## Set the handler.
		set bills_update_handler $HANDLER_ARG

		## Update the bills information on the client of this connection,
		## if there was an handler defined.
		if { $bills_update_handler != {} } { updateBills {} }
	}

	## Set the bill update handler:
	itcl::body Connection::setBillUpdateHandler { HANDLER_ARG } \
 	{
		## Set the handler.
		set bill_update_handler $HANDLER_ARG
	}

	## Set the login handler:
	itcl::body Connection::setLoginHandler { HANDLER_ARG } \
 	{
		set login_handler $HANDLER_ARG
	}

	## Set the users update handler:
	itcl::body Connection::setUsersUpdateHandler { HANDLER_ARG } \
 	{
		## Set the handler.
		set users_update_handler $HANDLER_ARG

		## Update the user information on the client of this connection,
		## if there was an handler defined.
		if { $users_update_handler != {} } updateUsers
	}

	## Update the article information on the client of this connection.
	itcl::body Connection::updateArticles {} \
	{
		## Go through all article objects in the cash register.
		set ARTICLE_INFO_LIST {}
		foreach ARTICLE [ $cash_register getArticleList ] \
		{
			## Get the configuration of each article object.
			lappend ARTICLE_INFO_LIST [ getObjectConfiguration $ARTICLE ] 
		}

		## Do some substitutions with the handler and send the result to the client.
		regsub -all {\\|&} [ list $ARTICLE_INFO_LIST ] {\\\0} SUBST
		regsub -all "%a" $articles_update_handler $SUBST COMMAND
		puts $fd $COMMAND
	}

	## Update a single bill information on the client of this connection.
	itcl::body Connection::updateBill { BILL_OBJECT_ARG } \
	{
		## Do some substitutions with the handler and send the result to the client.
		regsub -all {\\|&} [ list [ getObjectConfiguration $BILL_OBJECT_ARG ] ] {\\\0} SUBST 
		regsub -all "%b" $bill_update_handler $SUBST COMMAND
		puts $fd $COMMAND

	}

	## Update the bills information on the client of this connection.
	itcl::body Connection::updateBills { SEE_TAG_ARG } \
	{
		## Go through all bill objects in the cash register.
		set BILLS_INFO_LIST {}
		foreach BILL [ $cash_register cget -bills ] \
		{
			## Get the configuration of each bill object.
			set BILLS_INFO [ getObjectConfiguration $BILL ]

			## List only the bills of the current user.
 			if { [ lindex $BILLS_INFO [ expr [ lsearch -exact $BILLS_INFO "-user" ] +1 ] ] == $user } \
			{
				lappend BILLS_INFO_LIST $BILLS_INFO
			}
		}

		## Do some substitutions with the handler and send the result to the client.
		regsub -all {\\|&} [ list $BILLS_INFO_LIST ] {\\\0} SUBST
		regsub -all "%b" $bills_update_handler $SUBST COMMAND
		regsub -all {\\|&} [ list $SEE_TAG_ARG ] {\\\0} SUBST
		regsub -all "%s" $COMMAND $SUBST COMMAND
		puts $fd $COMMAND
	}

	## Update the user information on the client of this connection.
	itcl::body Connection::updateUsers {} \
	{
		## Go through all user objects for this server.
		set USER_INFO_LIST {}
		foreach USER [ $server getUserList ] \
		{
			## Get the configuration of each user object.
			set USER_INFO [ getObjectConfiguration $USER ]

			## Remove the password field.
 			set PASSWORD_INDEX [ lsearch -exact $USER_INFO "-password" ]
			set USER_INFO [ lreplace $USER_INFO $PASSWORD_INDEX [ expr $PASSWORD_INDEX +1 ] ]

			## Search the flags field for the "hide" flag.
 			if { [ lsearch -exact [ lindex $USER_INFO [ expr [ lsearch -exact $USER_INFO "-flags" ] +1 ] ] "hide" ] == -1 } \
			{
				## Not present. Add the user info record to the array.
				lappend USER_INFO_LIST $USER_INFO
			}
		}

		## Do some substitutions with the handler and send the result to the client.
		regsub -all {\\|&} [ list $USER_INFO_LIST ] {\\\0} SUBST
		regsub -all "%u" $users_update_handler $SUBST COMMAND
		puts $fd $COMMAND
	}
}

