#
# This script is part of TkKasse. You may use, modify and distribute it under the terms
# of the GNU General Public License, Version 2.
# See details at http://www.fsf.org/licenses/gpl.txt
# (C) by Jan Kandziora <tkkasse@users.sf.net> 
#


## Evaluate in the TkKasse namespace
namespace eval TkKasse \
{
	## Guest receipt printer filter class:
	itcl::class PrintFilter::GuestReceipt \
	{
		inherit TkKasse::PrintFilter

		constructor { args } { eval configure $args } {}

		method convert { DATA_ARG FILTER_ARG }

		public \
		{
			variable footer_text 
			variable header_text 
		}
	}

	## Convert data to print decription.
	itcl::body PrintFilter::GuestReceipt::convert { DATA_ARG FILTER_ARG } \
	{
		switch $FILTER_ARG \
		{
			GuestReceipt - GuestReceiptSpending - GuestReceiptRoom \
			{
				## Get access to the data.
				array set DATA_ARRAY $DATA_ARG

				set PRINT_DESCRIPTION [ list \
					[ list \
						filter \
						items \
						header_text \
						footer_text \
						tag_description \
						tag \
						table_description \
						table \
						date_description \
						date \
						date_format \
						tax_low_description \
						low_tax \
						tax_high_description \
						high_tax \
						gross_sum_low_tax_description \
						gross_sum_low_tax \
						gross_sum_high_tax_description \
						gross_sum_high_tax \
						sum_description \
						spending_text \
						room_text \
					] \
					[ list \
						$FILTER_ARG \
						$DATA_ARRAY(ITEMS) \
						$header_text \
						$footer_text \
						[ msgcat::mc <receipt.guest.tag.description> ] $DATA_ARRAY(TAG) \
						[ msgcat::mc <receipt.guest.table.description> ] $DATA_ARRAY(TABLE) \
						[ msgcat::mc <receipt.guest.date.description> ] $DATA_ARRAY(TIME) \
						[ msgcat::mc <receipt.guest.date.format> ] \
						[ msgcat::mc <receipt.guest.tax.description> $DATA_ARRAY(TAX_RATE_LOW) ] $DATA_ARRAY(LOW_TAX) \
						[ msgcat::mc <receipt.guest.tax.description> $DATA_ARRAY(TAX_RATE_HIGH) ] $DATA_ARRAY(HIGH_TAX) \
						[ msgcat::mc <receipt.guest.tax.low.description> ] $DATA_ARRAY(GROSS_SUM_LOW_TAX) \
						[ msgcat::mc <receipt.guest.tax.high.description> ] $DATA_ARRAY(GROSS_SUM_HIGH_TAX) \
						[ msgcat::mc <receipt.guest.sum.description> ] \
						[ msgcat::mc <receipt.guest.spending.description> ] \
						[ msgcat::mc <receipt.guest.room.description> ] \
					] \
				{
					## Compose the header.
					item::table header -columns {{0 1}} -slaves [ list \
						[ item::fill #auto -char "-" -height 1 ] \
						[ item::text #auto -sticky s -data $header_text ] \
						[ item::fill #auto -char "-" -height 1 ] ]

					## Compose the top of bill field.
					item::pad topofbill -slave \
						[ item::table #auto -columns {{1 1} {1 1}} -slaves [ list \
							[ item::string #auto -data $tag_description ] \
							[ item::string #auto -data $tag -position right ] \
							[ item::string #auto -data $table_description ] \
							[ item::string #auto -data $table -position right ] \
							[ item::string #auto -data $date_description ] \
							[ item::string #auto -data [ clock format $date -format $date_format ] -position right ] ] ] -pady 1

					## Compose the items field.
					set ITEM_SLAVES {}
					foreach ITEM $items \
					{
						## At first, the price of the item, after processing the extras, the price of the item plus extras.
						set PRICE [ lindex $ITEM 2 ]

						## Determine the multiplier.
						set MULTIPLIER [ lindex $ITEM 4 ]

						## Process the item itself.
						lappend ITEM_SLAVES [ item::string #auto -data [ lindex $ITEM 1 ] ]
						lappend ITEM_SLAVES [ item::fill #auto ] 
						if { $MULTIPLIER > 1 } { lappend ITEM_SLAVES [ item::string #auto -data "(" ] } { lappend ITEM_SLAVES [ item::fill #auto ] } 

						lappend ITEM_SLAVES [ item::string #auto -data [ formatCurrency $PRICE ] -position right ]
						if { $MULTIPLIER > 1 } { lappend ITEM_SLAVES [ item::string #auto -data ")" ] } { lappend ITEM_SLAVES [ item::fill #auto ] }

						## Process the extras, if any.
						foreach EXTRA [ lindex $ITEM 3 ] \
						{
							## Calculate the price.
							set EXTRA_PRICE [ lindex $EXTRA 1 ]
							set PRICE [ expr $PRICE + $EXTRA_PRICE ]

							## Process the extra.
							lappend ITEM_SLAVES [ item::pad #auto -slave [ item::string #auto -data [ lindex $EXTRA 0 ] ] -padx 1 ]
							lappend ITEM_SLAVES [ item::fill #auto ] 
							if { $MULTIPLIER > 1 } { lappend ITEM_SLAVES [ item::string #auto -data "(" ] } { lappend ITEM_SLAVES [ item::fill #auto ] } 
							lappend ITEM_SLAVES [ item::string #auto -data [ formatCurrency $EXTRA_PRICE ] -position right ]
							if { $MULTIPLIER > 1 } { lappend ITEM_SLAVES [ item::string #auto -data ")" ] } { lappend ITEM_SLAVES [ item::fill #auto ] }
						}

						## Print the multiplier and multiplied price, if the multiplicator is > 1.
						if { $MULTIPLIER > 1 } \
						{
							## Apply the multiplier to the item price.
							set PRICE [ expr $PRICE * $MULTIPLIER ]

							lappend ITEM_SLAVES [ item::string #auto -data "x $MULTIPLIER" -position center ]
							lappend ITEM_SLAVES [ item::fill #auto ] 
							lappend ITEM_SLAVES [ item::fill #auto ] 
							lappend ITEM_SLAVES [ item::string #auto -data [ formatCurrency $PRICE ] -position right ]
							lappend ITEM_SLAVES [ item::fill #auto ]
						}
					}
					item::table items -columns {{1 1} {1 0} {1 0} {6 0} {1 0}} -slaves $ITEM_SLAVES

					## Sum separator.
					item::fill separator -char "=" -height 1

					## Compose the sum field.
					set SUM_SLAVES {}
					if { $gross_sum_low_tax != 0 } \
					{
						lappend SUM_SLAVES [ item::string #auto -data $gross_sum_low_tax_description -position right ]
						lappend SUM_SLAVES [ item::pad #auto -slave [ item::string #auto -data [ getCurrencySymbol ] ] -padx 1 ] 
						lappend SUM_SLAVES [ item::string #auto -data [ formatCurrency $gross_sum_low_tax ] -position right ]
						lappend SUM_SLAVES [ item::fill #auto ]
						lappend SUM_SLAVES [ item::string #auto -data $tax_low_description -position right ]
						lappend SUM_SLAVES [ item::pad #auto -slave [ item::string #auto -data [ getCurrencySymbol ] ] -padx 1 ]
						lappend SUM_SLAVES [ item::string #auto -data [ formatCurrency $low_tax ] -position right ]
						lappend SUM_SLAVES [ item::fill #auto ]
					}
					if { $gross_sum_high_tax != 0 } \
					{
						lappend SUM_SLAVES [ item::string #auto -data $gross_sum_high_tax_description -position right ]
						lappend SUM_SLAVES [ item::pad #auto -slave [ item::string #auto -data [ getCurrencySymbol ] ] -padx 1 ] 
						lappend SUM_SLAVES [ item::string #auto -data [ formatCurrency $gross_sum_high_tax ] -position right ]
						lappend SUM_SLAVES [ item::fill #auto ]
						lappend SUM_SLAVES [ item::string #auto -data $tax_high_description -position right ]
						lappend SUM_SLAVES [ item::pad #auto -slave [ item::string #auto -data [ getCurrencySymbol ] ] -padx 1 ] 
						lappend SUM_SLAVES [ item::string #auto -data [ formatCurrency $high_tax ] -position right ]
						lappend SUM_SLAVES [ item::fill #auto ]
					}

					item::table sum -columns {{0 1}} -slaves [ list \
						[ item::table #auto -columns {{0 1} {0 0} {7 0} {1 0}} -slaves $SUM_SLAVES ] \
						[ item::pad #auto -slave [ item::table #auto -columns {{0 1} {0 0} {7 0} {1 0}} -slaves [ list \
							[ item::string #auto -data $sum_description -position right ] \
							[ item::pad #auto -slave [ item::string #auto -data [ getCurrencySymbol ] ] -padx 1 ] \
							[ item::string #auto -data [ formatCurrency [ expr $gross_sum_low_tax + $gross_sum_high_tax ] ] -position right ] \
							[ item::fill #auto ] ] ] -pady 1 ] ]

					## Compose the footer according to the receipt type.
					set FOOTER_SLAVES {}
					switch $filter \
					{
						GuestReceiptSpending \
						{
							lappend FOOTER_SLAVES [ item::fill #auto -char "-" -height 1 ]
							lappend FOOTER_SLAVES [ item::pad #auto -slave [ item::text #auto -sticky w -data $spending_text ] -pady 1 ] \
						}
						GuestReceiptRoom \
						{
							lappend FOOTER_SLAVES [ item::fill #auto -char "-" -height 1 ]
							lappend FOOTER_SLAVES [ item::pad #auto -slave [ item::text #auto -sticky w -data $room_text ] -pady 1 ] \
						}
					}
					lappend FOOTER_SLAVES [ item::fill #auto -char "-" -height 1 ]
					lappend FOOTER_SLAVES [ item::text #auto -sticky s -data $footer_text ]
					lappend FOOTER_SLAVES [ item::fill #auto -char "-" -height 1 ]
					item::table footer -columns {{0 1}} -slaves $FOOTER_SLAVES

					## Compose the receipt.
					item::table receipt \
						-columns {{1 1}} \
						-slaves {header topofbill items separator sum footer}

					print receipt
				} ]
			}
			default { return -code error -errorcode { TKKASSE PRINTER FILTER } "printer filter GuestReceipt not suitable for request" }
		}

		## Return the print description.
		return $PRINT_DESCRIPTION
	}



	## Order receipt printer filter class:
	itcl::class PrintFilter::OrderReceipt \
	{
		inherit TkKasse::PrintFilter

		constructor { args } { eval configure $args } {}

		method convert { DATA_ARG FILTER_ARG }
	}

	## Convert data to print decription.
	itcl::body PrintFilter::OrderReceipt::convert { DATA_ARG FILTER_ARG } \
	{
		if { $FILTER_ARG != "OrderReceipt" } { return -code error -errorcode { TKKASSE PRINTER FILTER } "printer filter OrderReceipt not suitable for request" }

		## Get access to the data.
		array set DATA_ARRAY $DATA_ARG

		set PRINT_DESCRIPTION [ list {tag table items} [ list $DATA_ARRAY(TAG) $DATA_ARRAY(TABLE) $DATA_ARRAY(ITEMS) ] \
		{
			## Compose the topofbill field.
			item::pad topofbill -slave [ item::table #auto -columns {{1 1} {6 0}} -slaves [ list [ item::string #auto -data $tag -position left ] [ item::string #auto -data $table -position right ] ] ] -pady 1

			## Compose the items field. Sort by the cancel field.
			set ITEM_SLAVES {}
			foreach ITEM [ lsort -index 4 -increasing $items ] \
			{
				## Determine the multiplier.
				set MULTIPLIER [ lindex $ITEM 3 ]

				## Process the item itself.
				lappend ITEM_SLAVES [ item::string #auto -data [ lindex $ITEM 0 ] ]
				lappend ITEM_SLAVES [ item::pad #auto -slave [ item::string #auto -data [ lindex $ITEM 1 ] ] -padx 1 ]

				## Add - for cancel orders.
				if [ lindex $ITEM 4 ] { lappend ITEM_SLAVES [ item::fill #auto -char "-" ] } { lappend ITEM_SLAVES [ item::fill #auto ] }

				## Process the extras, if any.
				foreach EXTRA [ lindex $ITEM 2 ] \
				{
					## Process the extra.
					lappend ITEM_SLAVES [ item::fill #auto ]
					lappend ITEM_SLAVES [ item::pad #auto -slave [ item::string #auto -data $EXTRA ] -padx 1 ]
					lappend ITEM_SLAVES [ item::fill #auto ]
				}

				## Print the multiplier and multiplied price, if the multiplicator is > 1.
				if { $MULTIPLIER > 1 } \
				{
					lappend ITEM_SLAVES [ item::fill #auto ]
					lappend ITEM_SLAVES [ item::string #auto -data "x $MULTIPLIER" -position center ]
					lappend ITEM_SLAVES [ item::fill #auto ]
				}
			}
			item::table items -columns {{5 0} {1 1} {3 0}} -slaves $ITEM_SLAVES

			## Compose the receipt.
			item::table receipt \
				-columns {{1 1}} \
				-slaves {topofbill items}

			print receipt
		} ]

		## Return the print description.
		return $PRINT_DESCRIPTION
	}



	## Alternative order receipt printer filter class:
	## For double size printing, meaning, only 20 chars/line!
	itcl::class PrintFilter::OrderReceiptDS \
	{
		inherit TkKasse::PrintFilter

		constructor { args } { eval configure $args } {}

		method convert { DATA_ARG FILTER_ARG }
	}

	## Convert data to print decription.
	itcl::body PrintFilter::OrderReceiptDS::convert { DATA_ARG FILTER_ARG } \
	{
		if { $FILTER_ARG != "OrderReceipt" } { return -code error -errorcode { TKKASSE PRINTER FILTER } "printer filter OrderReceipt not suitable for request" }

		## Get access to the data.
		array set DATA_ARRAY $DATA_ARG
		log debug $DATA_ARG
		set PRINT_DESCRIPTION [ list {tag table items} [ list $DATA_ARRAY(TAG) $DATA_ARRAY(TABLE) $DATA_ARRAY(ITEMS) ] \
		{
			## Compose the topofbill field.
			item::pad topofbill -slave [ item::table #auto -columns {{1 1} {6 0}} -slaves [ list [ item::string #auto -data $tag -position left ] [ item::string #auto -data $table -position right ] ] ] -pady 1

			## Compose the items field. Sort by the cancel field.
			set ITEM_SLAVES {}
			foreach ITEM [ lsort -index 4 -increasing $items ] \
			{
				## Determine the multiplier.
				set MULTIPLIER [ lindex $ITEM 3 ]

				## Process the item number.
				set ITEM_LINE [ item::string #auto -data [ lindex $ITEM 0 ] ]
				lappend ITEM_LINE [ item::fill #auto ]

				## Add - for cancel orders.
				if [ lindex $ITEM 4 ] { lappend ITEM_LINE [ item::fill #auto -char "-" ] } { lappend ITEM_LINE [ item::fill #auto ] }

				## Compose the item number and the first line.
				lappend ITEM_SLAVES [ item::table #auto -columns {{5 0} {1 1} {3 0}} -slaves $ITEM_LINE ]
				lappend ITEM_SLAVES [ item::string #auto -data [ lindex $ITEM 1 ] ]

				## Process the extras, if any.
				foreach EXTRA [ lindex $ITEM 2 ] \
				{
					## Process the extra.
					lappend ITEM_SLAVES [ item::string #auto -data $EXTRA ]
					lappend ITEM_SLAVES [ item::string #auto -data $EXTRA -position r ]
				}

				## Print the multiplier and multiplied price, if the multiplicator is > 1.
				if { $MULTIPLIER > 1 } \
				{
					lappend ITEM_SLAVES [ item::string #auto -data "x $MULTIPLIER" -position center ]
				}
			}

			## Compose the receipt.
			item::table receipt \
				-columns {{1 1}} \
				-slaves [ concat topofbill $ITEM_SLAVES ]

			print receipt
		} ]

		## Return the print description.
		return $PRINT_DESCRIPTION
	}



	## Cash register receipt printer filter class:
	itcl::class PrintFilter::CashRegisterReceipt \
	{
		inherit TkKasse::PrintFilter

		constructor { args } { eval configure $args } {}

		method convert { DATA_ARG FILTER_ARG }
	}

	## Convert data to print decription.
	itcl::body PrintFilter::CashRegisterReceipt::convert { DATA_ARG FILTER_ARG } \
	{
		if { $FILTER_ARG != "CashRegisterReceipt" } { return -code error -errorcode { TKKASSE PRINTER FILTER } "printer filter CashRegisterReceipt not suitable for request" }

		## Get access to the data.
		array set DATA_ARRAY $DATA_ARG

		set PRINT_DESCRIPTION [ list \
			[ list \
				bills \
				header_text \
				waiter_description \
				date_format \
				start_of_work_description \
				start_of_work \
				end_of_work_description \
				end_of_work \
				tag_description \
				payed_by_description \
				sum_description \
				payed_by_comment_description \
				tag_group_description \
				payed_by_short_description \
				payed_by_sum_description \
				sums \
				tax_description tax \
			] \
			[ list \
				$DATA_ARRAY(BILLS) \
				[ msgcat::mc <receipt.cash_register.header> ] \
				[ msgcat::mc <receipt.cash_register.waiter.description> ] \
				[ msgcat::mc <receipt.date.format> ] \
				[ msgcat::mc <receipt.work.start.description> ] $DATA_ARRAY(START_TIME) \
				[ msgcat::mc <receipt.work.end.description> ] $DATA_ARRAY(END_TIME) \
				[ msgcat::mc <receipt.tag.description> ] \
				[ msgcat::mc <receipt.payed_by.description> ] \
				[ msgcat::mc <receipt.sum.description> ] \
				[ msgcat::mc <receipt.payed_by.comment> ] \
				[ msgcat::mc <receipt.tag.group.description> ] \
				[ msgcat::mc <receipt.payed_by.short.description> ] \
				[ msgcat::mc <receipt.payed_by.sum.description> ] $DATA_ARRAY(SUMS) \
				[ msgcat::mc <receipt.tax.description> ] $DATA_ARRAY(TAX) \
			 ] \
		{
			## Compose the header.
			item::table header -columns {{0 1}} -slaves [ list \
				[ item::fill #auto -char "-" -height 1 ] \
				[ item::text #auto -sticky s -data $header_text ] \
				[ item::fill #auto -char "-" -height 1 ] ]

			## Compose the top of bill field.
			item::pad topofbill -slave \
				[ item::table #auto -columns {{1 1} {1 1}} -slaves [ list \
					[ item::string #auto -data $start_of_work_description ] \
					[ item::string #auto -data [ clock format $start_of_work -format $date_format ] -position right ] \
					[ item::string #auto -data $end_of_work_description ] \
					[ item::string #auto -data [ clock format $end_of_work -format $date_format ] -position right ] ] ] -pady 1

			## Compose the item header.
			item::table itemheader -columns {{0 1}} -slaves [ list \
				[ item::table #auto -columns {{4 0} {4 0} {7 0} {5 0} {0 1}} -slaves [ list \
					[ item::string #auto -data $tag_description -position center ] \
					[ item::pad #auto -slave [ item::string #auto -data $payed_by_description -position center ] -padx 1 ] \
					[ item::string #auto -data $sum_description -position center ] \
					[ item::pad #auto -slave [ item::string #auto -data $waiter_description -position center ] -padx 1 ] \
					[ item::string #auto -data $payed_by_comment_description -position center ] ] ] \
				[ item::fill #auto -char "-" -height 1 ] ] 

			## Compose the item list.
			set ITEMS {}
			set OLD_TAG_PREFIX {}
			foreach BILL $bills \
			{
				## Seperate the bill number prefix.
				## Get the last complete number in the counter string
				regexp {(.*?)/([0-9]+$)} [ lindex $BILL 0 ] RESULT TAG_PREFIX TAG_NUMBER

				## If the prefix has changed, insert an extra line.
				if { $TAG_PREFIX != $OLD_TAG_PREFIX } \
				{
					lappend ITEMS [ item::table #auto -columns {{1 1} {1 1}} -slaves [ list \
						[ item::string #auto -data $tag_group_description ] \
						[ item::string #auto -data $TAG_PREFIX -position right ] ] ]

					set OLD_TAG_PREFIX $TAG_PREFIX
				}

				## Get the sum.
				set SUM [ lindex $BILL 1 ]

				## Set the payed by short description array.
				array set PAYED_BY_SHORT_DESCRIPTION $payed_by_short_description

				## Append a item line.
				lappend ITEMS [ item::table #auto -columns {{4 0} {4 0} {7 0} {5 0} {0 1}} -slaves [ list \
					[ item::string #auto -data $TAG_NUMBER -position right ] \
					[ item::pad #auto -slave [ item::string #auto -data $PAYED_BY_SHORT_DESCRIPTION([ lindex $BILL 2 ]) -position center ] -padx 1 ] \
					[ item::string #auto -data [ formatCurrency $SUM ] -position right ] \
					[ item::pad #auto -slave [ item::string #auto -data [ lindex $BILL 4 ] -position center ] -padx 1 ] \
					[ item::string #auto -data [ lindex $BILL 3 ] -position right ] ] ] 

				## If the cancelled sum is other than zero, insert an extra line.
				set CANCELLED_SUM [ lindex $BILL 5 ]
				if { $CANCELLED_SUM != 0 } \
				{
					lappend ITEMS [ item::table #auto -columns {{4 0} {4 0} {7 0} {5 0} {0 1}} -slaves [ list \
						[ item::string #auto -data $TAG_NUMBER -position right ] \
						[ item::pad #auto -slave [ item::fill #auto -char "-" ] -padx 1 ] \
						[ item::string #auto -data [ formatCurrency $CANCELLED_SUM ] -position right ] \
						[ item::pad #auto -slave [ item::string #auto -data [ lindex $BILL 4 ] -position center ] -padx 1 ] \
						[ item::pad #auto -slave [ item::fill #auto -char "-" ] -padx 1 ] ] ] 
				}
			}
			item::table items -columns {{1 1}} -slaves [ concat $ITEMS [ item::fill #auto -char "-" -height 1 ] ]

			## Set the payed by sum description and sums array.
			array set PAYED_BY_SUM_DESCRIPTION $payed_by_sum_description
			array set SUMS $sums

			## Compose the sums field.
			set SUM_ITEMS {}
			foreach SUM [ array names SUMS ] \
			{
				lappend SUM_ITEMS [ item::string #auto -data $PAYED_BY_SUM_DESCRIPTION($SUM) -position left ]
				lappend SUM_ITEMS [ item::string #auto -data [ formatCurrency $SUMS($SUM) ] -position right ]
			}
			item::pad sums -slave [ item::table #auto -columns {{0 1} {8 0}} -slaves $SUM_ITEMS ] -pady 1 

			## Compose the tax field.
			item::pad tax -slave \
				[ item::table #auto -columns {{0 1} {7 0}} -slaves [ list \
					[ item::string #auto -data $tax_description ] \
					[ item::string #auto -data [ formatCurrency $tax ] -position right ] ] ]

			## Compose the footer.
			item::fill footer -char "-" -height 1

			## Compose the receipt.
			item::table receipt \
				-columns {{1 1}} \
				-slaves {header topofbill itemheader items sums tax footer}

			print receipt
		} ]

		## Return the print description.
		return $PRINT_DESCRIPTION
	}



	## Waiter receipt printer filter class:
	itcl::class PrintFilter::WaiterReceipt \
	{
		inherit TkKasse::PrintFilter

		constructor { args } { eval configure $args } {}

		method convert { DATA_ARG FILTER_ARG }
	}

	## Convert data to print decription.
	itcl::body PrintFilter::WaiterReceipt::convert { DATA_ARG FILTER_ARG } \
	{
		if { $FILTER_ARG != "WaiterReceipt" } { return -code error -errorcode { TKKASSE PRINTER FILTER } "printer filter WaiterReceipt not suitable for request" }

		## Get access to the data.
		array set DATA_ARRAY $DATA_ARG

		set PRINT_DESCRIPTION [ list \
			[ list \
				bills \
				header_text \
				waiter_description \
				waiter \
				date_format \
				start_of_work_description \
				start_of_work \
				end_of_work_description \
				end_of_work \
				tag_description \
				payed_by_description \
				sum_description \
				payed_by_comment_description \
				tag_group_description \
				payed_by_short_description \
				payed_by_sum_description \
				sums \
			] \
			[ list \
				$DATA_ARRAY(BILLS) \
				[ msgcat::mc <receipt.waiter.header> ] \
				[ msgcat::mc <receipt.waiter.waiter.description> ] \
				$DATA_ARRAY(USER) \
				[ msgcat::mc <receipt.date.format> ] \
				[ msgcat::mc <receipt.work.start.description> ] $DATA_ARRAY(START_TIME) \
				[ msgcat::mc <receipt.work.end.description> ] $DATA_ARRAY(END_TIME) \
				[ msgcat::mc <receipt.tag.description> ] \
				[ msgcat::mc <receipt.payed_by.description> ] \
				[ msgcat::mc <receipt.sum.description> ] \
				[ msgcat::mc <receipt.payed_by.comment> ] \
				[ msgcat::mc <receipt.tag.group.description> ] \
				[ msgcat::mc <receipt.payed_by.short.description> ] \
				[ msgcat::mc <receipt.payed_by.sum.description> ] $DATA_ARRAY(SUMS) \
			] \
		{
			## Compose the header.
			item::table header -columns {{0 1}} -slaves [ list \
				[ item::fill #auto -char "-" -height 1 ] \
				[ item::text #auto -sticky s -data $header_text ] \
				[ item::fill #auto -char "-" -height 1 ] ]

			## Compose the top of bill field.
			item::pad topofbill -slave \
				[ item::table #auto -columns {{1 1} {1 1}} -slaves [ list \
					[ item::string #auto -data $waiter_description ] \
					[ item::string #auto -data $waiter -position right ] \
					[ item::string #auto -data $start_of_work_description ] \
					[ item::string #auto -data [ clock format $start_of_work -format $date_format ] -position right ] \
					[ item::string #auto -data $end_of_work_description ] \
					[ item::string #auto -data [ clock format $end_of_work -format $date_format ] -position right ] ] ] -pady 1

			## Compose the item header.
			item::table itemheader -columns {{0 1}} -slaves [ list \
				[ item::table #auto -columns {{4 0} {4 0} {7 0} {0 1}} -slaves [ list \
					[ item::string #auto -data $tag_description -position center ] \
					[ item::pad #auto -slave [ item::string #auto -data $payed_by_description -position center ] -padx 1 ] \
					[ item::string #auto -data $sum_description -position center ] \
					[ item::pad #auto -slave [ item::string #auto -data $payed_by_comment_description -position center ] -padx 1 ] ] ] \
				[ item::fill #auto -char "-" -height 1 ] ]

			## Compose the item list.
			set ITEMS {}
			set OLD_TAG_PREFIX {}
			foreach BILL $bills \
			{
				## Seperate the bill number prefix.
				## Get the last complete number in the counter string
				regexp {(.*?)/([0-9]+$)} [ lindex $BILL 0 ] RESULT TAG_PREFIX TAG_NUMBER

				## If the prefix has changed, insert an extra line.
				if { $TAG_PREFIX != $OLD_TAG_PREFIX } \
				{
					lappend ITEMS [ item::table #auto -columns {{1 1} {1 1}} -slaves [ list \
						[ item::string #auto -data $tag_group_description ] \
						[ item::string #auto -data $TAG_PREFIX -position right ] ] ]

					set OLD_TAG_PREFIX $TAG_PREFIX
				}

				## Get the sum.
				set SUM [ lindex $BILL 1 ]

				## Set the payed by short description array.
				array set PAYED_BY_SHORT_DESCRIPTION $payed_by_short_description

				## Append a item line.
				lappend ITEMS [ item::table #auto -columns {{4 0} {4 0} {7 0} {0 1}} -slaves [ list \
					[ item::string #auto -data $TAG_NUMBER -position right ] \
					[ item::pad #auto -slave [ item::string #auto -data $PAYED_BY_SHORT_DESCRIPTION([ lindex $BILL 2 ]) -position center ] -padx 1 ] \
					[ item::string #auto -data [ formatCurrency $SUM ] -position right ] \
					[ item::pad #auto -slave [ item::string #auto -data [ lindex $BILL 3 ] -position right ] -padx 1 ] ] ] 

				## If the cancelled sum is other than zero, insert an extra line.
				set CANCELLED_SUM [ lindex $BILL 4 ]
				if { $CANCELLED_SUM != 0 } \
				{
					lappend ITEMS [ item::table #auto -columns {{4 0} {4 0} {7 0} {0 1}} -slaves [ list \
						[ item::string #auto -data $TAG_NUMBER -position right ] \
						[ item::pad #auto -slave [ item::fill #auto -char "-" ] -padx 1 ] \
						[ item::string #auto -data [ formatCurrency $CANCELLED_SUM ] -position right ] \
						[ item::pad #auto -slave [ item::fill #auto -char "-" ] -padx 1 ] ] ] 
				}
			}
			item::table items -columns {{1 1}} -slaves [ concat $ITEMS [ item::fill #auto -char "-" -height 1 ] ]

			## Set the payed by sum description and sums array.
			array set PAYED_BY_SUM_DESCRIPTION $payed_by_sum_description
			array set SUMS $sums

			## Compose the sums field.
			set SUM_ITEMS {}
			foreach SUM [ array names SUMS ] \
			{
				lappend SUM_ITEMS [ item::string #auto -data $PAYED_BY_SUM_DESCRIPTION($SUM) -position left ]
				lappend SUM_ITEMS [ item::string #auto -data [ formatCurrency $SUMS($SUM) ] -position right ]
			}
			item::pad sums -slave [ item::table #auto -columns {{0 1} {8 0}} -slaves $SUM_ITEMS ] -pady 1 

			## Compose the footer.
			item::fill footer -char "-" -height 1

			## Compose the receipt.
			item::table receipt \
				-columns {{1 1}} \
				-slaves {header topofbill itemheader items sums footer}

			print receipt
		} ]

		## Return the print description.
		return $PRINT_DESCRIPTION
	}
}

