#
# This script is part of TkKasse. You may use, modify and distribute it under the terms
# of the GNU General Public License, Version 2.
# See details at http://www.fsf.org/licenses/gpl.txt
# (C) by Jan Kandziora <tkkasse@users.sf.net> 
#


## Evaluate in the TkKasse namespace
namespace eval TkKasse \
{
	## TkKasse print filter class:
	itcl::class PrintFilter \
	{
		constructor { args } { eval configure $args } {}

		method convert { DATA_ARG FILTER_ARG }
	}



	## TkKasse printer driver class:
	itcl::class PrinterDriver \
	{
		constructor { args } { eval configure $args } {}
		
		method convert { PRINT_DESCRIPTION_ARG }
	}


	## TkKasse print spooler class:
	itcl::class PrintSpooler \
	{
		constructor { args } { eval configure $args } {}
		
		method print { PRINT_DATA_ARG }
	}



	## TkKasse printer class:
	itcl::class Printer \
	{
		constructor { args } { eval configure $args } {}

		method print { DATA_ARG FILTER_ARG DISPLAY_ARG SERVICE_ARG }

		public \
		{
			variable displays {}
			variable driver {}
			variable filters {}
			variable services {}
			variable spooler {}
		}
	}

	## Print the receipt.
	itcl::body Printer::print { DATA_ARG FILTER_ARG DISPLAY_ARG SERVICE_ARG } \
	{
		## Check the desired display and print location is matched.
		if  { ( [ lsearch -exact $displays $DISPLAY_ARG ] == -1 ) || ( [ lsearch -exact $services $SERVICE_ARG ] == -1 ) } \
		{
			## Printer doesn't match the display.
			return -code error -errorcode { TKKASSE PRINTER LOCATION } "printer location,service not suitable for request"
		}

		## Put the data into the filter.
		foreach FILTER $filters \
		{
			if [ catch { $FILTER convert $DATA_ARG $FILTER_ARG } PRINT_DESCRIPTION ] \
			{
				## The receipt could no be printed. Reason is given by the printer filter object.
				global errorCode errorInfo
				if { $errorCode != { TKKASSE PRINTER FILTER } } { log err "Print filter error: $errorInfo" }
			} \
			{
				## Success. Convert the print description and output the data.
				$spooler print [ $driver convert $PRINT_DESCRIPTION ] $DISPLAY_ARG $SERVICE_ARG

				## Success. Return without error.
				return
			}
		}

		## Printer doesn't match the filter.
		return -code error -errorcode { TKKASSE PRINTER FILTER } "printer filter not suitable for request"
	}



	## TkKasse print manager class:
	itcl::class PrintManager \
	{
		constructor { args } { eval configure $args } {}

		method print { DATA_ARG FILTER_ARG DISPLAY_ARG SERVICE_ARG }

		public \
		{
			variable printers {}
		}
	}

	## Print the receipt.
	itcl::body PrintManager::print { DATA_ARG FILTER_ARG DISPLAY_ARG SERVICE_ARG } \
	{
		log debug "Print attempt for $FILTER_ARG on $DISPLAY_ARG,$SERVICE_ARG."
		
		## Print on the first printer matched in the printer list.
		foreach PRINTER $printers \
		{
			if [ catch { $PRINTER print $DATA_ARG $FILTER_ARG $DISPLAY_ARG $SERVICE_ARG } RESULT ] \
			{
				## The receipt could no be printed on this printer. 
				log debug "$PRINTER: $RESULT"
			} \
			{
				## Success.
				log debug "$PRINTER: Success."
				return
			}
		}

		## Warn about missing printer definition.
		log warn "No printer defined for $FILTER_ARG on $DISPLAY_ARG,$SERVICE_ARG. Receipt was discarded."
	}
}

