#
# This script is part of TkKasse. You may use, modify and distribute it under the terms
# of the GNU General Public License, Version 2.
# See details at http://www.fsf.org/licenses/gpl.txt
# (C) by Jan Kandziora <tkkasse@users.sf.net> 
#


## User class:
itcl::class User \
{
	constructor { args  } { eval configure $args } {}
	
	method checkPassword { PASSWORD_ARG }

	public \
	{
		variable login {}
		variable password "*"
		variable rights {}
	}
}

## Check the user's password:
itcl::body User::checkPassword { PASSWORD_ARG } \
{
	## Check the password against the saved one.
	if { $password != [ crypt::password $PASSWORD_ARG $password ] } \
	{
		## Authentication failed.
		return -code error -errorcode { TKKASSE USER AUTHENTICATION_FAILURE } "authentication failed"
	}

	## Ok. The password is correct.
}



## User manager class:
itcl::class UserManager \
{
	constructor { args  } { eval configure $args } {}

	method authenticateUser { USER_LOGIN_NAME_ARG USER_PASSWORD_ARG }
	method addUser { USER_OBJECT_ARG }
	method removeUser { USER_OBJECT_ARG }

	public variable users {}
}

## Authenticate an user:
itcl::body UserManager::authenticateUser { USER_LOGIN_NAME_ARG USER_PASSWORD_ARG } \
{
	## Find an user object which matches the template.
	foreach USER_OBJECT $users \
	{
		## Is this the user object?
		if { [ $USER_OBJECT cget -login ] == $USER_LOGIN_NAME_ARG } \
		{
			## Check the password.
			$USER_OBJECT checkPassword $USER_PASSWORD_ARG

			## Success. Return the object descriptor.
			return $USER_OBJECT 
		}
	}

	## No such user. Authentication failed.
	return -code error -errorcode { TKKASSE USER AUTHENTICATION_FAILURE } "authentication failed"
}

## Add an user:
itcl::body UserManager::addUser { USER_OBJECT_ARG } \
{
	## Test whether the given user login name is unique.
	foreach USER_OBJECT $users \
	{
		if { [ $USER_OBJECT cget -login ] == [ $USER_OBJECT_ARG cget -login ] } \
		{
			## This user already exists in our list.
			return -code error -errorcode { TKKASSE USER EXISTS } "user already exists"
		}
	}

	## Add the user object descriptor to our list.
	lappend users [ namespace which $USER_OBJECT_ARG ]
}

## Remove an user from our list:
itcl::body UserManager::removeUser { USER_OBJECT_ARG } \
{
	## Find the user object for this login name.
	foreach USER_OBJECT $users \
	{
		## Is this the user object?
		if { [ $USER_OBJECT cget -login ] == [ $USER_OBJECT_ARG cget -login ] } \
		{
			## Yes. Take this user object's descriptor from the list.
			set INDEX [ lsearch -exact $users $USER_OBJECT ]
			set users [ lreplace $users $INDEX $INDEX ]

			## Return with success.
			return 
		}
	}

	## The user was not found. Throw an error.
	return -code error -errorcode { TKKASSE USER UNKNOWN } "user unknown."
}

