#
# This script is part of TkKasse. You may use, modify and distribute it under the terms
# of the GNU General Public License, Version 2.
# See details at http://www.fsf.org/licenses/gpl.txt
# (C) by Jan Kandziora <tkkasse@users.sf.net> 
#


## Generate log entries about missing translations.
proc msgcat::mcunknown {LOCALE STRING} \
{
	log debug "Translation missing for \"$STRING\", locale \"[ msgcat::mcpreferences ]\"."
	return $STRING
}

## Get the configuration of an object.
proc getObjectConfiguration { OBJECT_ARG } \
{
	## Isolate the options and their values.
	set OBJECT_OPTIONS {}
	foreach OPTION [ $OBJECT_ARG configure ] \
	{
		append OBJECT_OPTIONS "[ list [ lindex $OPTION 0 ] ] [ list [ lindex $OPTION 2 ] ] "
	}

	## Return the options and their values
	return $OBJECT_OPTIONS				
}

## Get the configuration of all objects of a class.
proc getConfiguration { args } \
{
	set COMMANDS {}

	## Process each given class. 
	foreach CLASS $args \
	{
		## Dump all objects of this class.
		foreach OBJECT [ lsort -dictionary [ itcl::find objects -isa $CLASS ] ] \
		{
			## Append this object's reconstruction conmmand.
			append COMMANDS "[ $OBJECT info class ] [ namespace which $OBJECT ] [ getObjectConfiguration $OBJECT ]\n"
		}
	}

	## Return all commands.
	return $COMMANDS
}

## Format currency values.
proc formatCurrency { PRICE_ARG } \
{
	## Get the localized currency format.
	set DECIMAL_WIDTH [ msgcat::mc <currency.format.decimals> ]
	set DECIMAL_SEPARATOR [ msgcat::mc <currency.format.decimalSeparator> ]
	set EYECATCH_WIDTH [ msgcat::mc <currency.format.eyecatch> ]
	set EYECATCH_SEPARATOR [ msgcat::mc <currency.format.eyecatchSeparator> ]

	## Delete leading minus and all leading zeroes from PRICE_ARG.
	set PRICE [ regsub {^-*0*} $PRICE_ARG {} ]
	if { $PRICE == {} } { set PRICE 0 }

	## Seperate the decimals from the non-broken part and format them.
	if { $DECIMAL_WIDTH > 0 } \
	{
		set DECIMALS [ format "$DECIMAL_SEPARATOR%0${DECIMAL_WIDTH}d" [ expr $PRICE % round(pow(10,$DECIMAL_WIDTH)) ] ]
	} \
	{
		set DECIMALS {}
	}
	
	## Get the value of the non-broken part.
	set PRICE_NONBROKEN [ expr $PRICE / round(pow(10,$DECIMAL_WIDTH)) ]

	## Insert eyecatches, if neccesary.
	set EYECATCHES {}
	set EYECATCH_COUNTER 0 
	for { set I [ string length $PRICE_NONBROKEN ] } { $I >= 0 } { set I [ expr $I-1 ] } \
	{
		set EYECATCHES "[ string index $PRICE_NONBROKEN $I ]$EYECATCHES"

		incr EYECATCH_COUNTER
		if { ($EYECATCH_COUNTER > $EYECATCH_WIDTH) && ( $I > 0 ) } \
		{
			set EYECATCH_COUNTER 1
			set EYECATCHES "$EYECATCH_SEPARATOR$EYECATCHES"
		}
	}

	## Return the formatted negative value, if the value is negative.
	if { $PRICE_ARG < 0 } { return "-$EYECATCHES$DECIMALS" } 

	## Return the formatted value.
	return " $EYECATCHES$DECIMALS"
}

proc getCurrencySymbol {} \
{
	msgcat::mc <currency.symbol>
}
